#!/usr/bin/env python3
# -*- coding: utf-8 -*-
__copyright__ = """ This code is licensed under the 3-clause BSD license.
Copyright ETH Zurich, Department of Chemistry and Applied Biosciences, Reiher Group.
See LICENSE.txt for details.
"""
"""
Provides the ElectronicDataWidget class.
"""

from concurrent.futures import ThreadPoolExecutor
from typing import Any, Optional

from vtk import (
    vtkRenderer,
    vtkActor,
    vtkMarchingCubes,
    vtkPolyDataMapper,
    vtkImageData,
)

from scine_heron.electronic_data.electronic_data_image_generator import (
    ElectronicDataImageGenerator,
)
from scine_heron.electronic_data.electronic_data_status_manager import (
    ElectronicDataStatusManager,
)
from scine_heron.electronic_data.electronic_data import ElectronicData
from scine_heron.electronic_data.molden_file_reader import MoldenFileReader
from scine_heron.settings.settings_status_manager import SettingsStatusManager
from scine_heron.utilities import write_error_message


class ElectronicDataWidget:
    """
    Displays electronic data in a 3D view.
    """

    def __init__(
        self, renderer: vtkRenderer, settings_status_manager: SettingsStatusManager, mol_widget: Any
    ):
        self.electronic_data_status_manager = ElectronicDataStatusManager()
        self.electronic_data_status_manager.molden_input_changed_signal.connect(
            self.read_molden_input
        )
        self.electronic_data_status_manager.hamiltonian_changed.connect(
            self.clear_surfaces_and_data
        )

        self.__renderer = renderer
        self.__mo_plus_actor: vtkActor = None
        self.__mo_minus_actor: vtkActor = None
        self.__pool = ThreadPoolExecutor(2)

        # read electronic data from Molden file
        self.__reader = MoldenFileReader()
        self.__settings_status_manager = settings_status_manager
        self.__electronic_data: Optional[ElectronicData] = None
        self.__data_image_generator: Optional[ElectronicDataImageGenerator] = None
        self.__settings_status_manager.selected_mo_changed.connect(
            self.view_molecular_orbital
        )
        self.__settings_status_manager.hamiltonian_changed.connect(
            self.clear_surfaces_and_data
        )

        self.__mol_widget = mol_widget

    def get_electronic_data(self):
        return self.__electronic_data

    def read_molden_input(self, molden_input: str) -> None:
        self.__clear_isosurface()
        self.__electronic_data = self.__reader.read_molden(molden_input)
        self.__data_image_generator = ElectronicDataImageGenerator(
            self.__electronic_data
        )

        self.__settings_status_manager.number_of_molecular_orbital = len(
            self.__electronic_data.mo
        )

    # Do not use this function if you do not use autocas
    # I was/still am just to stupid to understand this
    # Will fix it in the future
    def view_orbital(self, mo_index):
        if self.__data_image_generator is None and mo_index == 0:
            return
        if self.__data_image_generator is not None:
            self.__settings_status_manager.selected_molecular_orbital = mo_index

    def view_molecular_orbital(self, molecular_orbital_index: int) -> None:
        if self.__data_image_generator is None and molecular_orbital_index == 0:
            return
        if molecular_orbital_index == 0:
            # reset MO data
            self.clear_surfaces_and_data()
        else:
            if self.__data_image_generator is None:
                result = self.__mol_widget.single_point_calculation()
                if not result:
                    return
                if not result.molden_input:
                    write_error_message("Chosen calculator does not support MO output")
                    return
                self.read_molden_input(result.molden_input)
            self.__clear_isosurface()
            if molecular_orbital_index == -1:  # HOMO
                self.__settings_status_manager.selected_molecular_orbital = self.__get_actual_index(
                    molecular_orbital_index
                )
                # we update to actual value and return to ensure that we have the actual number in the GUI
                return
            elif molecular_orbital_index == -2:  # LUMO
                # we update to actual value and return to ensure that we have the actual number in the GUI
                self.__settings_status_manager.selected_molecular_orbital = self.__get_actual_index(
                    molecular_orbital_index
                )
                return
            elif molecular_orbital_index == -3:
                msg = "The electron density is being calculated ..."
            else:
                msg = (
                    "The molecular orbital number "
                    + str(molecular_orbital_index)
                    + " is being calculated ..."
                )
            if self.__data_image_generator is not None:
                self.__settings_status_manager.info_message = msg
                future = self.__pool.submit(
                    self.__data_image_generator.generate_mo_image,
                    molecular_orbital_index - 1,
                )
                future.add_done_callback(self.__show_mo)

    def __get_actual_index(self, orbital_index: int) -> int:
        if self.__electronic_data is None:
            raise RuntimeError("Electronic Data has not been setup properly")
        i = 0
        while self.__electronic_data.mo[i].occupation >= 1.0:
            i += 1
        return i if orbital_index == -1 else i + 1

    def __show_mo(self, calculation_result) -> None:
        """
        Show molecular orbital with index "molecular_orbital_index" as an isosurface
        """
        if not self.__settings_status_manager.selected_molecular_orbital:
            return
        image = calculation_result.result()
        self.__create_isosurface(
            image, self.__settings_status_manager.molecular_orbital_value
        )
        self.__create_isosurface(
            image, -self.__settings_status_manager.molecular_orbital_value
        )
        self.__renderer.GetRenderWindow().Render()
        self.__settings_status_manager.info_message = (
            "The molecular orbital has been calculated."
        )
        return

    def __create_actor(
        self, mapper: vtkPolyDataMapper, value: float = 0.05
    ) -> vtkActor:
        """
        Create actor and set colors and opacity for it.
        """
        actor = vtkActor()
        actor.GetProperty().SetOpacity(0.5)
        actor_property = actor.GetProperty()

        # TODO dynamic style sheet colors
        if value < 0:
            actor_property.SetColor(1, 0.2, 0.2)  # red
        elif value > 0:
            actor_property.SetColor(0.2, 0.2, 1)  # blue

        actor.SetMapper(mapper)

        return actor

    def __create_isosurface(self, image: vtkImageData, value: float) -> None:
        iso_sphere = vtkMarchingCubes()
        iso_sphere.SetInputData(image)
        iso_sphere.ComputeNormalsOn()
        iso_sphere.GenerateValues(1, value, value)
        iso_sphere.Update()

        # Create a mapper and actor
        mapper = vtkPolyDataMapper()
        mapper.SetInputConnection(iso_sphere.GetOutputPort())
        mapper.ScalarVisibilityOff()
        if value > 0:
            if self.__mo_plus_actor is not None:
                self.__renderer.RemoveActor(self.__mo_plus_actor)

            self.__mo_plus_actor = self.__create_actor(mapper, value)
            self.__renderer.AddActor(self.__mo_plus_actor)
        else:
            if self.__mo_minus_actor is not None:
                self.__renderer.RemoveActor(self.__mo_minus_actor)

            self.__mo_minus_actor = self.__create_actor(mapper, value)
            self.__renderer.AddActor(self.__mo_minus_actor)

    def __clear_isosurface(self) -> None:
        self.__renderer.RemoveActor(self.__mo_plus_actor)
        self.__renderer.RemoveActor(self.__mo_minus_actor)

        self.__mo_plus_actor = None
        self.__mo_minus_actor = None

        self.__renderer.GetRenderWindow().Render()

    def clear_surfaces_and_data(self) -> None:
        self.__electronic_data = None
        self.__data_image_generator = None
        self.__clear_isosurface()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
__copyright__ = """ This code is licensed under the 3-clause BSD license.
Copyright ETH Zurich, Department of Chemistry and Applied Biosciences, Reiher Group.
See LICENSE.txt for details.
"""
from typing import Union

import scine_database as db


def get_compound_or_flask(object_id: db.ID, object_type: db.CompoundOrFlask, compounds: db.Collection,
                          flasks: db.Collection) -> Union[db.Compound, db.Flask]:
    """
    Construct the compound or flask object depending on the type. Flask and Compound have a large set of functions
    in common. Through this function, we do not have to differentiate between them at every point.

    Parameters
    ----------
    object_id : db.ID
        The ID of the object to construct.
    object_type : db.CompoundOrFlask
        The label for Compound or Flaks.
    compounds : db.Collection
        The compounds collection.
    flasks : db.Collection
        The flasks collection.

    Returns
    -------
    Either the flask or compound object.

    Note
    ----
    Raises a runtime error if the object_type is unknown.
    """
    if object_type == db.CompoundOrFlask.COMPOUND:
        return db.Compound(object_id, compounds)
    if object_type == db.CompoundOrFlask.FLASK:
        return db.Flask(object_id, flasks)
    raise RuntimeError("Requested aggregate type is not supported.")



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
__copyright__ = """ This code is licensed under the 3-clause BSD license.
Copyright ETH Zurich, Department of Chemistry and Applied Biosciences, Reiher Group.
See LICENSE.txt for details.
"""

from abc import abstractmethod
from typing import Any, Optional, Dict

from PySide2.QtWidgets import (
    QWidget,
    QVBoxLayout,
    QHBoxLayout,
    QCompleter,
    QToolButton,
    QSizePolicy,
    QLineEdit,
    QLabel,
    QPushButton,
)
from PySide2.QtCore import Qt, QObject, QSize

from scine_heron.utilities import write_error_message


class AbstractList(QWidget):  # typing help

    @abstractmethod
    def set_current_item(self, item: Any) -> None:
        pass

    def focus(self) -> None:
        pass


class CollapsibleListItem(QWidget):

    def __init__(self, parent: AbstractList, name: str, content_widget: QWidget) -> None:
        super().__init__(parent)
        self._parent = parent
        self.name = name

        self.toggle_button = QToolButton(parent=self)
        self.toggle_button.setText(name)
        self.toggle_button.setCheckable(True)
        self.toggle_button.setChecked(False)
        self.toggle_button.setStyleSheet("QToolButton { border: none; }")
        self.toggle_button.setToolButtonStyle(Qt.ToolButtonTextBesideIcon)
        self.toggle_button.setArrowType(Qt.RightArrow)
        self.toggle_button.pressed.connect(self.on_pressed)  # pylint: disable=no-member

        self.content = content_widget
        self.content.setVisible(False)
        self.content.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Expanding)

        layout = QVBoxLayout()
        layout.setSpacing(0)
        layout.setContentsMargins(0, 0, 0, 0)
        layout.addWidget(self.toggle_button)
        layout.addWidget(self.content)
        layout.setAlignment(Qt.AlignTop)
        self.setLayout(layout)
        self.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Expanding)

    def set_visibility(self, visible: bool) -> None:
        self.toggle_button.setChecked(visible)
        self.on_pressed()

    def on_pressed(self):
        self.content.setVisible(not self.content.isVisible())
        self.toggle_button.setArrowType(Qt.DownArrow if self.content.isVisible() else Qt.RightArrow)
        self.content.updateGeometry()
        self.updateGeometry()

    def size(self) -> QSize:
        content_size = self.content.size().height() if not self.toggle_button.isChecked() else 0
        return self.toggle_button.size() + QSize(0, content_size)


class CollapsibleList(AbstractList):

    def __init__(self, parent: QObject):
        super().__init__(parent)
        self._parent = parent
        self._layout = QVBoxLayout()
        self._layout.setAlignment(Qt.AlignTop)

        self._search_button = QPushButton("Ok")
        self._search_button.clicked.connect(  # pylint: disable=no-member
            self._search_and_display
        )
        self._search_box = QLineEdit("")
        self._search_box.returnPressed.connect(  # pylint: disable=no-member
            self._search_button.click
        )

        search_layout = QHBoxLayout()
        search_layout.addWidget(QLabel("Jump to:"))
        search_layout.addWidget(self._search_box)
        search_layout.addWidget(self._search_button)
        self._layout.addLayout(search_layout)

        self._widgets: Dict[str, CollapsibleListItem] = {}
        self._current_widget: Optional[str] = None
        self.setLayout(self._layout)
        self.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Expanding)

    def clear(self) -> None:
        for widget in self._widgets.values():
            self._layout.removeWidget(widget)
            widget.deleteLater()
        self._widgets = {}
        self._current_widget = None

    def _search_and_display(self) -> None:
        query = self._search_box.text().strip()
        if query not in self._widgets:
            write_error_message(f"The item '{query}' does not exist in the list")
        else:
            self.set_current_item(self._widgets[query])
            self.focus()

    def set_current_item(self, item: CollapsibleListItem) -> None:
        self._current_widget = item.name

    def focus(self):
        self._widgets[self._current_widget].set_visibility(True)
        if hasattr(self._parent, "focus"):
            self._parent.focus()
        self.updateGeometry()

    def add_widget(self, item: CollapsibleListItem) -> None:
        self._layout.addStretch(1)
        self._layout.addWidget(item)
        self._widgets[item.name] = item
        self._update_suggestions()

    def remove_widget(self, item: CollapsibleListItem) -> None:
        self._layout.removeWidget(item)
        item.deleteLater()
        self._widgets.pop(item.name)
        self._layout.removeWidget(item)
        item.deleteLater()
        self._update_suggestions()

    def _update_suggestions(self):
        completer = QCompleter(list(self._widgets.keys()))
        self._search_box.setCompleter(completer)

    def get_item_by_name(self, name: str) -> Optional[CollapsibleListItem]:
        try:
            return self._widgets[name]
        except KeyError:
            return None

    def current_item(self) -> Optional[CollapsibleListItem]:
        if self._current_widget is None:
            return None
        return self._widgets[self._current_widget]

    def contextMenuEvent(self, event):
        current = self.current_item()
        if current is not None:
            current.contextMenuEvent(event)

    def __len__(self):
        return len(self._widgets)

    def __iter__(self):
        return (w for w in self._widgets.values())



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
__copyright__ = """ This code is licensed under the 3-clause BSD license.
Copyright ETH Zurich, Department of Chemistry and Applied Biosciences, Reiher Group.
See LICENSE.txt for details.
"""

from datetime import datetime
from typing import Any, Optional, Union, Tuple, List, Dict
from json import dumps

from PySide2.QtWidgets import (QWidget, QVBoxLayout, QHBoxLayout, QTabWidget, QMenu,
                               QTreeWidget, QTreeWidgetItem, QLabel, QLineEdit, QPushButton, QDateTimeEdit, QCheckBox)
from PySide2.QtCore import Qt, QObject, QDate
from PySide2.QtGui import QFont

import scine_utilities as su
from scine_database import Manager, ID, Structure, ElementaryStep
import scine_database as db
from scine_database.queries import optimized_labels_enums
from scine_database.concentration_query_functions import query_reaction_flux
from scine_database.energy_query_functions import check_barrier_height, get_single_barrier_for_elementary_step
from scine_utilities import AtomCollection

from scine_heron.utilities import datetime_to_query
from scine_heron.containers.buttons import TextPushButton
from scine_heron.molecule.molecule_widget import MoleculeWidget
from scine_heron.database.reaction_compound_widget import AdvancedSettingsWidget
from scine_heron.database.compound_and_flasks_helper import get_compound_or_flask
from scine_heron.styling.delegates import CustomLightDelegate
from scine_heron.utilities import copy_text_to_clipboard, write_error_message, write_info_message
from scine_heron import get_core_tab
import scine_heron.config as config
from scine_heron.dependencies.optional_import import importer, is_imported
JsonSerialization = importer("scine_molassembler", "JsonSerialization")
AggregateFilterBuilderButton = importer("scine_heron.chemoton.filter_builder", "AggregateFilterBuilderButton")
DefaultAggregateFilter = importer("scine_chemoton.filters.aggregate_filters", "AggregateFilter")


class TreeWidget(QTreeWidget):
    def __init__(self, parent: QObject):
        QTreeWidget.__init__(self, parent)
        if config.MODE in config.LIGHT_MODES:
            self.setItemDelegate(CustomLightDelegate())

    def contextMenuEvent(self, event):
        self.currentItem().contextMenuEvent(event)


class TreeWidgetItem(QTreeWidgetItem):

    def __init__(self, parent: QObject, info: List[str]):
        QTreeWidgetItem.__init__(self, parent, info, type=QTreeWidgetItem.UserType)
        self.id_string_list = [info[1]]
        self.item_type = info[0]

    def contextMenuEvent(self, event) -> None:
        menu = QMenu()
        copyid_action = menu.addAction('Copy ID')
        show_in_mol_view = menu.addAction('Move to Molecular Viewer')

        cfs_type: bool = self.item_type in ['Compound', 'Flask', 'Structure']
        er_type: bool = self.item_type in ['Elementary Step', 'Reaction']

        if cfs_type:
            copy_masm_cbor_action = menu.addAction('Copy Molassembler Graph')
            center_in_network = menu.addAction('Center in Network View')
        if er_type:
            submenu = menu.addMenu("Copy Reactants")
            reactants = self.get_reactants()
            reactant_copy_actions = [submenu.addAction(r) for r in reactants]
        _ = menu.addAction('Close Menu')
        action = menu.exec_(event.globalPos())  # type: ignore
        if action == copyid_action:
            current_id = self.id_string_list[0]
            copy_text_to_clipboard(current_id)
        if action == show_in_mol_view:
            self.display_in_molecule_viewer()
        if cfs_type:
            if action == copy_masm_cbor_action:
                self.copy_cbor_to_clipboard()
            if action == center_in_network:
                self.center_network_view_on_aggregate()
        if er_type:
            for i, copy_action in enumerate(reactant_copy_actions):
                if action == copy_action:
                    copy_text_to_clipboard(reactants[i])

    def copy_cbor_to_clipboard(self):
        if self.item_type == 'Structure':
            self.parent().copy_cbor_to_clipboard()
        else:
            self.treeWidget().parent().copy_cbor_to_clipboard(self.id_string_list[0], self.item_type)

    def center_network_view_on_aggregate(self):
        if self.item_type == 'Structure':
            self.parent().center_network_view_on_aggregate()
        else:
            self.treeWidget().parent().center_network_view_on_aggregate(self.id_string_list[0])

    def display_in_molecule_viewer(self):
        if self.item_type in ['Structure', 'Elementary Step']:
            self.treeWidget().parent().display_in_molecule_viewer(self.id_string_list[0])
        else:
            self.child(0).display_in_molecule_viewer()

    def get_reactants(self) -> List[str]:
        if self.item_type in ['Elementary Step']:
            return self.treeWidget().parent().get_reactants(self.id_string_list[0])
        else:
            return self.child(0).get_reactants()


class AggregateTreeWidget(QWidget):
    def __init__(self, parent: Optional[QWidget], db_manager: Manager):
        super(AggregateTreeWidget, self).__init__(parent)

        layout = QVBoxLayout()

        self.current_agg_id_label = QLabel(self)
        self.current_agg_id_label.resize(560, 40)
        self.current_agg_id_label.setText("Current Aggregate ID")
        layout.addWidget(self.current_agg_id_label)

        agg_widget = QWidget()
        agg_h_box = QHBoxLayout()
        self.current_agg_id_text = QLineEdit(self)
        self.current_agg_id_text.resize(400, 40)
        self.current_agg_id_text.setText("")
        self.current_agg_id_text.returnPressed.connect(self.__focus_aggregate_id)  # pylint: disable=no-member
        agg_h_box.addWidget(self.current_agg_id_text)
        self.button_jump_to_agg_id = QPushButton("Jump To ID")
        agg_h_box.addWidget(self.button_jump_to_agg_id)
        self.button_jump_to_agg_id.clicked.connect(self.__focus_aggregate_id)  # pylint: disable=no-member
        agg_widget.setLayout(agg_h_box)
        layout.addWidget(agg_widget)

        self.current_id_label = QLabel(self)
        self.current_id_label.resize(560, 40)
        self.current_id_label.setText("Current Structure ID")
        layout.addWidget(self.current_id_label)

        str_widget = QWidget()
        str_h_box = QHBoxLayout()
        self.current_id_text = QLineEdit(self)
        self.current_id_text.resize(400, 40)
        self.current_id_text.setText("")
        str_h_box.addWidget(self.current_id_text)
        self.button_jump_to_str_id = QPushButton("Jump To ID")
        str_h_box.addWidget(self.button_jump_to_str_id)
        self.current_id_text.returnPressed.connect(self.__focus_structure_id)  # pylint: disable=no-member
        self.button_jump_to_str_id.clicked.connect(self.__focus_structure_id)  # pylint: disable=no-member
        str_widget.setLayout(str_h_box)
        layout.addWidget(str_widget)

        self.current_cbor_label = QLabel(self)
        self.current_cbor_label.resize(560, 40)
        self.current_cbor_label.setText("Current Aggregate MASM-CBOR String")
        layout.addWidget(self.current_cbor_label)

        cbor_widget = QWidget()
        cbor_h_box = QHBoxLayout()
        self.current_cbor_text = QLineEdit(self)
        self.current_cbor_text.resize(400, 40)
        self.current_cbor_text.setText("")
        self.current_cbor_text.returnPressed.connect(self.jump_to_cbor)  # pylint: disable=no-member
        cbor_h_box.addWidget(self.current_cbor_text)
        self.button_jump_to_cbor = QPushButton("Jump To CBOR-String")
        cbor_h_box.addWidget(self.button_jump_to_cbor)
        self.button_jump_to_cbor.clicked.connect(self.jump_to_cbor)  # pylint: disable=no-member
        cbor_widget.setLayout(cbor_h_box)
        layout.addWidget(cbor_widget)

        update_h_box = QHBoxLayout()
        update_widget = QWidget()
        self.button_update = TextPushButton("Update", self.update, shortcut="F5")
        update_h_box.addWidget(self.button_update)
        self._query_compounds = True
        self.query_compounds_cbox = QCheckBox("Show Compounds", parent=self)
        self.query_compounds_cbox.setChecked(self._query_compounds)
        self.query_compounds_cbox.clicked.connect(self._check_query_compounds)  # pylint: disable=no-member
        update_h_box.addWidget(self.query_compounds_cbox)
        self.query_flasks_cbox = QCheckBox("Show Flasks", parent=self)
        self.query_flasks_cbox.setChecked(False)
        self.query_flasks_cbox.clicked.connect(self._check_query_flasks)  # pylint: disable=no-member
        update_h_box.addWidget(self.query_flasks_cbox)
        if is_imported(AggregateFilterBuilderButton):
            self.filter_button = AggregateFilterBuilderButton(parent=self, shortcut="Ctrl+F")
            update_h_box.addWidget(self.filter_button)
        else:
            self.filter_button = None
        update_widget.setLayout(update_h_box)
        layout.addWidget(update_widget)

        self.db_manager = db_manager

        self.tree = TreeWidget(self)
        self.tree.setColumnCount(1)
        self.tree.setHeaderHidden(True)
        self.tree.itemClicked.connect(self.focus)  # pylint: disable=no-member
        self.tree.itemActivated.connect(self.focus)  # pylint: disable=no-member
        layout.addWidget(self.tree)
        self.setLayout(layout)

        self.structures = self.db_manager.get_collection("structures")

        self.compounds = self.db_manager.get_collection("compounds")
        self.flasks = self.db_manager.get_collection("flasks")
        self._is_updating = False
        self._was_updated = False

    def _check_query_compounds(self):
        if self.query_compounds_cbox.isChecked():
            self.query_flasks_cbox.setChecked(False)
        else:
            self.query_flasks_cbox.setChecked(True)

    def _check_query_flasks(self):
        if self.query_flasks_cbox.isChecked():
            self.query_compounds_cbox.setChecked(False)
        else:
            self.query_compounds_cbox.setChecked(True)

    def _check_masm_graph(self, aggregate: Union[db.Compound, db.Flask]) -> bool:
        structure = db.Structure(aggregate.get_centroid(), self.structures)
        if not structure.has_graph("masm_cbor_graph"):
            return False
        if is_imported(JsonSerialization):
            return JsonSerialization.equal_molecules(structure.get_graph("masm_cbor_graph"),
                                                     self.current_cbor_text.text())
        return structure.get_graph("masm_cbor_graph") == self.current_cbor_text.text()

    def jump_to_cbor(self):
        if not self._was_updated:
            self.update()
        query_compounds = self.query_compounds_cbox.isChecked()
        if query_compounds:
            collection = self.compounds
            it = collection.iterate_all_compounds()
        else:
            collection = self.flasks
            it = collection.iterate_all_flasks()
        write_info_message("Searching...")
        for i in it:
            i.link(collection)
            if self._check_masm_graph(i):
                item_search = str(i.id()), None
                self._activate_item(item_search)
                break
        else:
            write_error_message("CBOR string was not found")
            return
        write_info_message("Found entry")

    def update(self):
        if self._is_updating:
            return
        self._is_updating = True
        self.tree.clear()
        items = []
        query_compounds = self.query_compounds_cbox.isChecked()
        self.tree.setColumnCount(2)
        if query_compounds:
            aggregates = "compounds"
            collection = self.compounds
            it = collection.iterate_all_compounds()
        else:
            aggregates = "flasks"
            collection = self.flasks
            it = collection.iterate_all_flasks()
        allowed_labels = optimized_labels_enums()
        agg_filter = None if self.filter_button is None else self.filter_button.get_aggregate_filter()
        have_to_filter: bool = agg_filter is not None and agg_filter is not DefaultAggregateFilter
        if have_to_filter:
            agg_filter.initialize_collections(self.db_manager)
        for i in it:
            i.link(collection)
            if have_to_filter and not agg_filter.filter(i):
                continue
            main = TreeWidgetItem(self.tree, [f"{aggregates.capitalize()[0:-1]}", str(i.get_id())])
            items.append(main)
            for sid in i.get_structures():
                structure = db.Structure(sid, self.structures)
                if structure.get_label() in allowed_labels:
                    TreeWidgetItem(main, ['Structure', str(sid)])
        # Fixed column size for first colum,
        #   because dynamic fit does not account for leaf sizes
        self.tree.setColumnWidth(0, 150)
        self.tree.resizeColumnToContents(1)
        self.tree.insertTopLevelItems(0, items)
        self._is_updating = False
        self._was_updated = True

    def focus(self, item: TreeWidgetItem, _):
        if item.childCount() > 0:
            self.current_agg_id_text.setText(item.id_string_list[0])
            item = item.child(0)
        else:
            self.current_agg_id_text.setText(item.parent().id_string_list[0])
        structure_id = item.id_string_list[0]
        structure = db.Structure(db.ID(structure_id), self.structures)
        cbor_string = ""
        if structure.has_graph("masm_cbor_graph"):
            cbor_string = structure.get_graph("masm_cbor_graph")
        self.current_cbor_text.setText(cbor_string)
        self.current_id_text.setText(structure_id)
        self.__focus_structure_id(structure_id)

    def __focus_structure_id(self, str_id: Optional[str] = None) -> None:
        if not self._was_updated:
            write_info_message("Gather database data first")
            self.update()
        if str_id is None or not str_id:
            str_id = self.current_id_text.text().strip()
            if not str_id or not self.__id_sanity_check(str_id):
                return
            write_info_message("Searching...")
            item_search = None, str_id
            self._activate_item(item_search)
        s = Structure(ID(str_id), self.structures)
        self.parent().display_molecule(s.get_atoms(), s.get_charge(), s.get_multiplicity())

    def __focus_aggregate_id(self, agg_id: Optional[str] = None) -> None:
        if not self._was_updated:
            write_info_message("Gather database data first")
            self.update()
        if agg_id is None or not agg_id:
            agg_id = self.current_agg_id_text.text().strip()
        if not agg_id or not self.__id_sanity_check(agg_id):
            return
        if self.query_compounds_cbox.isChecked():
            aggregates = "compounds"
            a_type = db.CompoundOrFlask.COMPOUND
        else:
            aggregates = "flasks"
            a_type = db.CompoundOrFlask.FLASK
        write_info_message("Searching...")
        agg = get_compound_or_flask(db.ID(agg_id), a_type, self.compounds, self.flasks)
        if not agg.exists():
            write_error_message(f"No {aggregates.capitalize()[:-1]} exists with ID {agg_id}")
            return
        str_id = str(agg.get_centroid())
        item_search = agg_id, None
        self._activate_item(item_search)
        s = Structure(ID(str_id), self.structures)
        self.parent().display_molecule(s.get_atoms(), s.get_charge(), s.get_multiplicity())

    @staticmethod
    def __id_sanity_check(id_: str) -> bool:
        try:
            _ = db.ID(id_)
            return True
        except RuntimeError:
            write_error_message(f"The entry '{id_}' is not a valid database ID")
            return False

    def _activate_item(self, item_search: Tuple[Union[str, None], Union[str, None]]) -> None:
        if item_search[0] is None and item_search[1] is None:
            return
        if item_search[0] is not None:
            items = self.tree.findItems(item_search[0], Qt.MatchFlag.MatchCaseSensitive, column=1)  # type: ignore
            if not items:
                write_error_message(f"Entry '{item_search[0]}' was not found")
                return
        if item_search[1] is not None:
            items = self.tree.findItems(item_search[1], Qt.MatchFlag.MatchCaseSensitive, column=1)  # type: ignore
            if not items:
                write_error_message(f"Entry '{item_search[1]}' was not found")
                return
        current = self.tree.currentItem()
        if current is not None:
            current.setSelected(False)
            current.setExpanded(False)
        self.tree.setCurrentItem(items[0])
        self.tree.scrollToItem(items[0])
        items[0].setExpanded(True)
        items[0].setSelected(True)
        self.focus(items[0], None)

    def copy_cbor_to_clipboard(self, aggregate_id: str, aggregate_type: str):
        if aggregate_type.lower() == "compound":
            a_type = db.CompoundOrFlask.COMPOUND
        else:
            a_type = db.CompoundOrFlask.FLASK
        aggregate = get_compound_or_flask(db.ID(aggregate_id), a_type, self.compounds, self.flasks)
        centroid = db.Structure(aggregate.get_centroid(), self.structures)
        cbor = centroid.get_graph("masm_cbor_graph")
        copy_text_to_clipboard(cbor)

    def center_network_view_on_aggregate(self, aggregate_id: str):
        tab = get_core_tab('network_viewer')
        if tab is not None:
            tab.center_on_aggregate(aggregate_id)

    def display_in_molecule_viewer(self, structure_id: str):
        structure = db.Structure(db.ID(structure_id), self.structures)
        atoms = structure.get_atoms()
        tab = get_core_tab('molecule_viewer')
        if tab is not None:
            tab.update_molecule(atoms=atoms)


class ReactionTreeWidget(QWidget):
    def __init__(self, parent: Optional[QWidget], db_manager: Manager):
        super(ReactionTreeWidget, self).__init__(parent)
        self.db_manager = db_manager

        self.collection = self.db_manager.get_collection("reactions")
        self.elementary_steps = self.db_manager.get_collection("elementary_steps")
        self.structures = self.db_manager.get_collection("structures")
        self.properties = self.db_manager.get_collection("properties")
        self.compounds = self.db_manager.get_collection("compounds")
        self.flasks = self.db_manager.get_collection("flasks")
        self.reactions = self.db_manager.get_collection("reactions")

        layout = QVBoxLayout()

        # Selection label and line edit box
        self.reaction_selection_label = QLabel(self)
        self.reaction_selection_label.setText('Current Reaction ID')
        self.reaction_selection_label.resize(560, 40)
        layout.addWidget(self.reaction_selection_label)

        self.reaction_selection_text = QLineEdit(self)
        self.reaction_selection_text.setText("")
        self.reaction_selection_text.setToolTip('Any valid reaction ID\n'
                                                ' e.g.: "627ad4ef38b7073df406d2ab"\n'
                                                'or a file name if <Reaction Ids from file> is toggled.\n'
                                                'Note that the file must only contain reaction IDs with one ID per '
                                                'line.')
        self.reaction_selection_text.resize(560, 40)
        layout.addWidget(self.reaction_selection_text)

        qh_box_layout_labeles = QHBoxLayout()
        self.lhs_selection_label = QLabel(self)
        self.lhs_selection_label.setText("LHS")
        self.lhs_selection_label.resize(int(560 / 2.1), 40)
        qh_box_layout_labeles.addWidget(self.lhs_selection_label)

        self.rhs_selection_label = QLabel(self)
        self.rhs_selection_label.setText("RHS")
        self.rhs_selection_label.resize(int(560 / 2.1), 40)
        qh_box_layout_labeles.addWidget(self.rhs_selection_label)
        layout.addLayout(qh_box_layout_labeles)

        qh_box_layout_texts = QHBoxLayout()
        self.lhs_selection_text = QLineEdit(self)
        self.lhs_selection_text.setText("")
        self.lhs_selection_text.setToolTip("The LHS/RHS constraint may be combined with the selection field if\n"
                                           "it contains a <$and> operator or reads reaction IDs from file.\n"
                                           "Multiple aggregate IDs/MASM graphs should be separated by spaces:\n"
                                           "    627a75f938b7073df306c7c3 62aa892438b70728a545f6e0")
        qh_box_layout_texts.addWidget(self.lhs_selection_text)

        self.rhs_selection_text = QLineEdit(self)
        self.rhs_selection_text.setText("")
        qh_box_layout_texts.addWidget(self.rhs_selection_text)
        layout.addLayout(qh_box_layout_texts)

        qh_box_layout_toggles = QHBoxLayout()

        # optional toggle to read the reaction ids from some file
        self.read_from_file_cbox = QCheckBox("Reaction IDs from file", parent=self)
        self.read_from_file_cbox.setChecked(False)
        qh_box_layout_toggles.addWidget(self.read_from_file_cbox)

        self.use_lhs_rhs_selection_cbox = QCheckBox("Use LHS/RHS selection", parent=self)
        self.use_lhs_rhs_selection_cbox.setChecked(False)
        qh_box_layout_toggles.addWidget(self.use_lhs_rhs_selection_cbox)

        self.take_as_masm_string_cbox = QCheckBox("LHS and RHS are MASM strings", parent=self)
        self.take_as_masm_string_cbox.setChecked(False)
        qh_box_layout_toggles.addWidget(self.take_as_masm_string_cbox)

        self.time_label = QLabel(self)
        self.time_label.resize(100, 40)
        self.time_label.setText("Only Show Modified Reactions Since")
        self.time_edit = QDateTimeEdit(QDate())
        self.time_edit.setDisplayFormat("HH:mm dd.MM.yyyy")
        qh_box_layout_toggles.addWidget(self.time_label)
        qh_box_layout_toggles.addWidget(self.time_edit)
        self.earliest_reaction_creation: Optional[datetime] = None

        layout.addLayout(qh_box_layout_toggles)

        qh_box_update_and_sort = QHBoxLayout()
        self.button_update = TextPushButton("Update", self.update, shortcut="F5")
        qh_box_update_and_sort.addWidget(self.button_update)

        self.sort_by_barrier = QCheckBox("Sort by Reaction Barrier Height", parent=self)
        self.sort_by_barrier.setChecked(False)
        qh_box_update_and_sort.addWidget(self.sort_by_barrier)
        layout.addLayout(qh_box_update_and_sort)

        if is_imported(AggregateFilterBuilderButton):
            self.filter_button = AggregateFilterBuilderButton(parent=self, shortcut="Ctrl+F")
            qh_box_update_and_sort.addWidget(self.filter_button)
            self.filter_both_sides_check_box = QCheckBox("Filter both sides", parent=self)
            self.filter_both_sides_check_box.setChecked(False)
            self.filter_both_sides_check_box.setToolTip("If checked, the filter will be applied to both sides of the "
                                                        "reaction.\nIf unchecked, it suffices if one side passes the "
                                                        "filter.")
            self.filter_all_reactants_check_box = QCheckBox("Filter all reactants per side", parent=self)
            self.filter_all_reactants_check_box.setChecked(False)
            self.filter_all_reactants_check_box.setToolTip(
                "If checked, the filter will be applied to all reactants per "
                "side of the reaction.\nIf unchecked, it suffices if one "
                "reactant per side passes the filter.")
            qh_box_update_and_sort.addWidget(self.filter_both_sides_check_box)
            qh_box_update_and_sort.addWidget(self.filter_all_reactants_check_box)
        else:
            self.filter_button = None

        self._advanced_settings_visible = False
        self._set_up_advanced_settings_widgets(layout)

        self.tree = TreeWidget(self)
        self.tree.setColumnCount(1)
        self.tree.setHeaderHidden(True)
        self.tree.itemClicked.connect(self.focus)  # pylint: disable=no-member
        self.tree.itemActivated.connect(self.focus)  # pylint: disable=no-member
        layout.addWidget(self.tree)
        self.setLayout(layout)

        self._is_updating = False
        self._was_updated = False

    def set_advanced_settings_visible(self) -> None:
        self._advanced_settings_visible = self.advanced_settings_cbox.isChecked()
        if self._advanced_settings_visible:
            self.advanced_settings_widget.setVisible(True)
        else:
            self.advanced_settings_widget.setVisible(False)

    def _set_up_advanced_settings_widgets(self, layout):
        self.advanced_settings_cbox = QCheckBox("ADVANCED SETTINGS")
        self.advanced_settings_cbox.setChecked(self._advanced_settings_visible)
        self.advanced_settings_cbox.toggled.connect(  # pylint: disable=no-member
            self.set_advanced_settings_visible
        )

        self.advanced_settings_widget = AdvancedSettingsWidget(
            None,
            self.db_manager
        )
        layout.addWidget(self.advanced_settings_cbox)
        layout.addWidget(self.advanced_settings_widget)
        self.set_advanced_settings_visible()

    def _get_reaction_selection(self) -> Dict[str, Any]:
        from json import JSONDecodeError
        selection: Dict = {"$and": []}
        custom_selection: Optional[Dict] = None

        if self.read_from_file_cbox.isChecked():
            try:
                f = open(self.reaction_selection_text.text(), "r")
                lines = f.readlines()
                f.close()
                reaction_str_ids = [{"$oid": line.replace("\n", "")} for line in lines]
                custom_selection = {"_id": {"$in": reaction_str_ids}}
            except OSError:
                write_error_message("Could not open/read file " + self.reaction_selection_text.text())
        else:
            if self.reaction_selection_text.text():
                try:
                    # TODO: id only is enough here, don't make it complicated
                    custom_selection = {"_id": {"$oid": self.reaction_selection_text.text()}}
                except JSONDecodeError as error:
                    write_error_message(
                        "Invalid input string. The string must be convertable to a Python dictionary!")
                    write_error_message(error.msg)
        if custom_selection is not None:
            selection["$and"].append(custom_selection)

        # time handling
        time = datetime.fromtimestamp(self.time_edit.dateTime().toSecsSinceEpoch())
        if self.earliest_reaction_creation is None:
            reaction = self.reactions.get_one_reaction(dumps({}))
            if reaction is None:
                # we don't even have a reaction, so just return empty dict
                return dict()
            self.earliest_reaction_creation = reaction.created()
        if self.earliest_reaction_creation < time:
            # only query if even relevant for network
            selection["$and"].append(datetime_to_query(time))

        # Use LHS/RHS definition as an additional constraint
        if self.use_lhs_rhs_selection_cbox.isChecked() and not self.take_as_masm_string_cbox.isChecked():
            lhs_selection_ids = self.lhs_selection_text.text().split()
            rhs_selection_ids = self.rhs_selection_text.text().split()
            if lhs_selection_ids:
                lhs_condition = {"$and": [{"lhs.id": {"$oid": str_id}} for str_id in lhs_selection_ids]}
                selection["$and"].append(lhs_condition)  # type: ignore
            if rhs_selection_ids:
                rhs_condition = {"$and": [{"rhs.id": {"$oid": str_id}} for str_id in rhs_selection_ids]}
                selection["$and"].append(rhs_condition)  # type: ignore

        if selection["$and"]:
            return selection
        return dict()

    def update(self):
        if self._is_updating:
            return
        self._is_updating = True
        items = []
        self.tree.setSortingEnabled(False)
        self.tree.setColumnCount(3)
        selection = self._get_reaction_selection()
        self.advanced_settings_widget.update_settings()
        self.tree.clear()
        agg_filter = None if self.filter_button is None else self.filter_button.get_aggregate_filter()
        have_to_filter: bool = agg_filter is not None and agg_filter is not DefaultAggregateFilter
        if have_to_filter:
            agg_filter.initialize_collections(self.db_manager)
        for i in self.collection.iterate_reactions(dumps(selection)):
            i.link(self.collection)
            if not self._filter_reaction(i):
                continue
            if have_to_filter:
                lhs, rhs = i.get_reactants(db.Side.BOTH)
                lhs_types, rhs_types = i.get_reactant_types(db.Side.BOTH)
                lhs_is_ok = bool(
                    all(agg_filter.filter(get_compound_or_flask(lid, ltype, self.compounds, self.flasks))
                        for lid, ltype in zip(lhs, lhs_types))
                    if self.filter_all_reactants_check_box.isChecked()
                    else any(agg_filter.filter(get_compound_or_flask(lid, ltype, self.compounds, self.flasks))
                             for lid, ltype in zip(lhs, lhs_types))
                )
                if not lhs_is_ok and self.filter_both_sides_check_box.isChecked():
                    continue
                rhs_is_ok = bool(
                    all(agg_filter.filter(get_compound_or_flask(rid, rtype, self.compounds, self.flasks))
                        for rid, rtype in zip(rhs, rhs_types))
                    if self.filter_all_reactants_check_box.isChecked()
                    else any(agg_filter.filter(get_compound_or_flask(rid, rtype, self.compounds, self.flasks))
                             for rid, rtype in zip(rhs, rhs_types))
                )
                if not lhs_is_ok and not rhs_is_ok:
                    continue
                if not rhs_is_ok and self.filter_both_sides_check_box.isChecked():
                    continue

            main = TreeWidgetItem(self.tree, ['Reaction', str(i.get_id())])
            items.append(main)
            if self.sort_by_barrier.isChecked():
                min_barrier = None
            for e in i.get_elementary_steps():
                leaf = TreeWidgetItem(main, ['Elementary Step', str(e)])
                if self.sort_by_barrier.isChecked():
                    step = db.ElementaryStep(e)
                    step.link(self.elementary_steps)
                    barrier = get_single_barrier_for_elementary_step(
                        step,
                        self.advanced_settings_widget.get_model(),
                        self.structures,
                        self.properties
                    )
                    if barrier is None:
                        leaf.setText(2, 'None')
                    else:
                        leaf.setText(2, f'{barrier:-6.1f}')
                        if min_barrier is None or min_barrier > barrier:
                            min_barrier = barrier
            if self.sort_by_barrier.isChecked():
                if min_barrier is not None:
                    main.setText(2, f'{min_barrier:-6.1f}')
                else:
                    main.setText(2, 'None')
        if self.sort_by_barrier.isChecked():
            self.tree.insertTopLevelItems(0, items)
            self.tree.setSortingEnabled(True)
            self.tree.sortItems(2, Qt.AscendingOrder)
            self.tree.resizeColumnToContents(2)
        else:
            self.tree.insertTopLevelItems(0, items)
        # Fixed column size for first colum,
        #   because dynamic fit does not account for leaf sizes
        self.tree.setColumnWidth(0, 200)
        self.tree.resizeColumnToContents(1)
        self._is_updating = False
        self._was_updated = False

    def _filter_reaction(self, reaction: db.Reaction) -> bool:
        if not self._check_reaction_barrier(reaction):
            return False
        if self.advanced_settings_widget.scale_with_concentrations():
            if not self._check_concentration_flux(reaction):
                return False
        if self.take_as_masm_string_cbox.isChecked() and self.use_lhs_rhs_selection_cbox.isChecked():
            if not self._check_masm_graph(reaction):
                return False
        return True

    def _check_reaction_barrier(self, reaction: db.Reaction) -> bool:
        # TODO: replace by correct database call
        return check_barrier_height(reaction, self.db_manager, self.advanced_settings_widget.get_model(),
                                    self.structures, self.properties, self.advanced_settings_widget.get_max_barrier(),
                                    self.advanced_settings_widget.get_min_barrier(),
                                    self.advanced_settings_widget.always_show_barrierless()) is not None

    def _check_concentration_flux(self, reaction: db.Reaction) -> bool:
        flux = query_reaction_flux("_reaction_edge_flux", reaction, self.compounds, self.flasks, self.structures,
                                   self.properties)
        return flux > self.advanced_settings_widget.get_min_flux()

    def _valid_graphs(self, reference: List[str], reactant_graphs: List[str]) -> bool:
        # Make sure that there are no graphs but the reference ones.
        for graph in reactant_graphs:
            if graph not in reference:
                return False
        # Make sure that we have all reference graphs.
        for ref in reference:
            if ref not in reactant_graphs:
                return False
        return True

    def _get_reactant_masm_strings(self, reactants: List[db.ID], reactant_types: List[db.CompoundOrFlask]) -> List[str]:
        masm_strings = list()
        for a_id, a_type in zip(reactants, reactant_types):
            aggregate = get_compound_or_flask(a_id, a_type, self.compounds, self.flasks)
            centroid = db.Structure(aggregate.get_centroid(), self.structures)
            graph = centroid.get_graph("masm_cbor_graph")
            masm_strings.append(graph)
        return masm_strings

    def _check_masm_graph(self, reaction: db.Reaction) -> bool:
        reactants = reaction.get_reactants(db.Side.BOTH)
        reactant_types = reaction.get_reactant_types(db.Side.BOTH)
        lhs_references = self.lhs_selection_text.text().split()
        rhs_references = self.rhs_selection_text.text().split()
        reaction_lhs_masm_strings = self._get_reactant_masm_strings(reactants[0], reactant_types[0])
        reaction_rhs_masm_strings = self._get_reactant_masm_strings(reactants[1], reactant_types[1])
        valid_lhs_strings = True
        valid_rhs_strings = True
        if lhs_references:
            valid_lhs_strings = self._valid_graphs(lhs_references, reaction_lhs_masm_strings)
        if rhs_references:
            valid_rhs_strings = self._valid_graphs(rhs_references, reaction_rhs_masm_strings)

        if valid_rhs_strings and valid_lhs_strings:
            return True
        return valid_rhs_strings and valid_lhs_strings

    def focus(self, item, _):
        if item.childCount() > 0:
            item = item.child(0)
        step = ElementaryStep(ID(item.text(1)))
        step.link(self.elementary_steps)
        if not step.has_transition_state():
            ts_atoms = AtomCollection()
        else:
            ts = Structure(step.get_transition_state())
            ts.link(self.structures)
            ts_atoms = ts.get_atoms()
        lhs, rhs = step.get_reactants(db.Side.BOTH)
        lhs_atoms = []
        for lid in lhs:
            s = Structure(lid)
            s.link(self.structures)
            lhs_atoms.append(s.get_atoms())
        rhs_atoms = []
        for rid in rhs:
            s = Structure(rid)
            s.link(self.structures)
            rhs_atoms.append(s.get_atoms())
        self.parent().display_molecules(lhs_atoms, ts_atoms, rhs_atoms)

    def display_in_molecule_viewer(self, step_id: str):
        step = db.ElementaryStep(db.ID(step_id), self.elementary_steps)
        if not step.has_transition_state():
            return
        else:
            ts_id = step.get_transition_state()
            structure = db.Structure(ts_id, self.structures)
            atoms = structure.get_atoms()
            tab = get_core_tab('molecule_viewer')
            if tab is not None:
                tab.update_molecule(atoms=atoms)

    def get_reactants(self, step_id: str) -> List[str]:
        step = db.ElementaryStep(db.ID(step_id), self.elementary_steps)
        reaction = db.Reaction(step.get_reaction(), self.reactions)
        reactants = reaction.get_reactants(db.Side.BOTH)
        # return flattened list
        return [str(r_id) for side in reactants for r_id in side]


class MolecularReactionView(QTabWidget):
    def __init__(self, parent: Optional[QWidget], lhs, ts, rhs, mol_widget_cache):
        super(MolecularReactionView, self).__init__(parent)
        self.__layout = None
        self.__deleteables: List[QWidget] = []
        self.update(lhs, ts, rhs, mol_widget_cache)

    def update(self, lhs, ts, rhs, mol_widget_cache):
        if self.__layout is None:
            self.__layout = QHBoxLayout()
            self.setLayout(self.__layout)
        else:
            for i in reversed(range(self.__layout.count())):
                widget = self.__layout.itemAt(i).widget()
                widget.setParent(None)  # actually remove the widget from the gui
                self.__layout.removeWidget(widget)
            for d in self.__deleteables:
                d.close()

        def plus():
            plus = QLabel(self)
            plus.setText("+")
            plus.setFont(QFont('Arial', 20))
            return plus

        arrow_lhs = QLabel(self)
        arrow_lhs.setText("---")
        arrow_lhs.setFont(QFont('Arial', 20))
        arrow_rhs = QLabel(self)
        arrow_rhs.setFont(QFont('Arial', 20))
        arrow_rhs.setText("-->")

        mol_view_count = 0
        for i, atoms in enumerate(lhs):
            mol_widget = mol_widget_cache[mol_view_count]
            mol_widget.update_molecule(atoms=atoms)
            self.__layout.addWidget(mol_widget)
            if i < len(lhs) - 1:
                plus_widget = plus()
                self.__layout.addWidget(plus_widget)
                self.__deleteables.append(plus_widget)
            mol_view_count += 1
        self.__layout.addWidget(arrow_lhs)
        self.__deleteables.append(arrow_lhs)
        mol_widget = mol_widget_cache[mol_view_count]
        mol_widget.update_molecule(atoms=ts)
        mol_view_count += 1
        self.__layout.addWidget(mol_widget)
        self.__layout.addWidget(arrow_rhs)
        self.__deleteables.append(arrow_rhs)
        for i, atoms in enumerate(rhs):
            mol_widget = mol_widget_cache[mol_view_count]
            mol_widget.update_molecule(atoms=atoms)
            self.__layout.addWidget(mol_widget)
            if i < len(rhs) - 1:
                plus_widget = plus()
                self.__layout.addWidget(plus_widget)
                self.__deleteables.append(plus_widget)
            mol_view_count += 1

        while mol_view_count < len(mol_widget_cache):
            mol_widget = mol_widget_cache[mol_view_count]
            mol_widget.update_molecule(atoms=lhs[0])
            mol_view_count += 1


class ContentWidget(QWidget):
    def __init__(self, parent: Optional[QWidget], db_manager: Manager, aggregates: str):
        super(ContentWidget, self).__init__(parent)
        self.db_manager = db_manager
        self.__layout = QVBoxLayout()

        if aggregates == "reactions":
            self.tree = ReactionTreeWidget(self, db_manager)
        else:
            self.tree = AggregateTreeWidget(self, db_manager)
        self.__layout.addWidget(self.tree)
        self.__mol_widget_cache: List[MoleculeWidget] = [
            MoleculeWidget(
                parent=self, atoms=AtomCollection(), disable_modification=True
            )
        ]
        self.mol_widget = self.__mol_widget_cache[0]
        self.__layout.addWidget(self.mol_widget)
        self.setLayout(self.__layout)

    def display_molecule(self, atoms: su.AtomCollection,
                         charge: Optional[int] = None, multiplicity: Optional[int] = None) -> None:
        self.__layout.removeWidget(self.mol_widget)
        self.mol_widget = self.__mol_widget_cache[0]
        self.mol_widget.update_molecule(atoms=atoms)
        tooltip = ""
        if charge is not None:
            tooltip += f"Charge: {charge}"
        if multiplicity is not None:
            if tooltip:
                tooltip += "; "
            tooltip += f"Multiplicity: {multiplicity}"
        if tooltip:
            self.mol_widget.setToolTip(tooltip)
        self.__layout.addWidget(self.mol_widget)

    def display_molecules(self, lhs, ts, rhs):
        n_views = len(lhs) + 1 + len(rhs)
        for _ in range(n_views - len(self.__mol_widget_cache)):
            self.__mol_widget_cache.append(MoleculeWidget(
                parent=self, atoms=AtomCollection(), disable_modification=True
            ))
        if isinstance(self.mol_widget, MoleculeWidget):
            self.__layout.removeWidget(self.mol_widget)
            self.mol_widget = MolecularReactionView(self, lhs, ts, rhs, self.__mol_widget_cache)
            self.__layout.addWidget(self.mol_widget)
        elif isinstance(self.mol_widget, MolecularReactionView):
            self.mol_widget.update(lhs, ts, rhs, self.__mol_widget_cache)


class DatabaseViewerWidget(QWidget):
    def __init__(self, parent: Optional[QWidget], db_manager: Manager):
        super(DatabaseViewerWidget, self).__init__(parent)
        self.db_manager = db_manager

        self.__layout = QVBoxLayout()
        self.tabs = QTabWidget()
        self.aggregates_tab = ContentWidget(self, self.db_manager, "compounds")
        self.reactions_tab = ContentWidget(self, self.db_manager, "reactions")
        # Add tabs
        self.tabs.addTab(self.aggregates_tab, "Aggregates")
        self.tabs.addTab(self.reactions_tab, "Reactions")
        # Add tabs to widget
        self.__layout.addWidget(self.tabs)
        self.setLayout(self.__layout)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
__copyright__ = """ This code is licensed under the 3-clause BSD license.
Copyright ETH Zurich, Department of Chemistry and Applied Biosciences, Reiher Group.
See LICENSE.txt for details.
"""

from typing import Optional
from PySide2.QtCore import QObject
from PySide2.QtWidgets import (
    QDialog,
    QLabel
)

from scine_chemoton.gears.kinetic_modeling.atomization import (
    MultiModelEnergyReferences,
    PlaceHolderMultiModelEnergyReferences
)

from scine_heron.containers.layouts import VerticalLayout


class EnergyReferenceBuilder(QDialog):
    def __init__(self, _: MultiModelEnergyReferences, parent: Optional[QObject] = None):
        """
        Builder for a MultiModelEnergyReferences object. This is a rather complicated object that can also be
        constructed by the kinetic modeling gear through a keyword argument. Therefore, we do not support its
        construction directly.
        """
        super().__init__(parent=parent)
        # widgets
        self._label = QLabel("Advanced option to set specific atomization energies.\n"
                             "This feature is not supported through the GUI.\n"
                             "Energies are referenced according to the option\n"
                             "<energy_reference_type>.")
        # layout
        layout = VerticalLayout([self._label])
        self.setLayout(layout)

    def get_energy_reference(self) -> MultiModelEnergyReferences:
        return PlaceHolderMultiModelEnergyReferences()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
__copyright__ = """ This code is licensed under the 3-clause BSD license.
Copyright ETH Zurich, Department of Chemistry and Applied Biosciences, Reiher Group.
See LICENSE.txt for details.
"""
from vtk import (
    vtkPolyData,
    vtkAlgorithm,
    vtkSphereSource,
    vtkPolyDataMapper,
    vtkActor,
    vtkInformationVector,
)
from vtk.util.vtkAlgorithm import VTKPythonAlgorithmBase
from scine_heron.haptic.haptic_pointer_data import HapticPointerData

from typing import Any, List


class PointToSphereAlgorithm(VTKPythonAlgorithmBase):  # type: ignore[misc]
    """
    vtkAlgorithm that can be used to convert the haptic pointer
    to a sphere source.
    """

    def __init__(self) -> None:
        super().__init__(
            nInputPorts=1,
            nOutputPorts=1,
            inputType=["vtkPolyData"],
            outputType="vtkPolyData",
        )

    def FillInputPortInformation(self, port: int, info: Any) -> int:
        """Sets the required input type to InputType."""
        info.Set(vtkAlgorithm.INPUT_REQUIRED_DATA_TYPE(), self.InputType[port])
        return 1

    def RequestData(
        self,
        _: Any,
        in_info: List[vtkInformationVector],
        out_info: vtkInformationVector,
    ) -> int:
        """
        Performs the transformation
        """
        in_position = vtkPolyData.GetData(in_info[0]).GetPoints().GetPoint(0)

        out_sphere = vtkPolyData.GetData(out_info)

        sphere = vtkSphereSource()
        sphere.SetCenter(*in_position)
        sphere.SetRadius(0.25)
        sphere.SetPhiResolution(25)
        sphere.SetThetaResolution(25)
        sphere.Update()
        out_sphere.DeepCopy(sphere.GetOutput())

        return 1


class HapticPointerRepresentation:
    def __init__(self, haptic_pointer_data: HapticPointerData) -> None:
        self._haptic_pointer_data = haptic_pointer_data
        self._haptic_pointer_source = PointToSphereAlgorithm()
        self._haptic_pointer_mapper = vtkPolyDataMapper()
        self._haptic_pointer_actor = vtkActor()

        # connecting everything
        self._haptic_pointer_source.SetInputConnection(haptic_pointer_data.output)
        self._haptic_pointer_mapper.SetInputConnection(
            self._haptic_pointer_source.GetOutputPort()
        )
        self._haptic_pointer_actor.SetMapper(self._haptic_pointer_mapper)
        self._haptic_pointer_actor.GetProperty().SetColor(0, 0.75, 0.75)

    @property
    def actor(self) -> vtkActor:
        return self._haptic_pointer_actor

    def update_pointer_position(self, pos: Any) -> None:
        self._haptic_pointer_data.update_pointer_position(pos)

    @property
    def source(self) -> vtkSphereSource:
        return self._haptic_pointer_source



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
__copyright__ = """ This code is licensed under the 3-clause BSD license.
Copyright ETH Zurich, Department of Chemistry and Applied Biosciences, Reiher Group.
See LICENSE.txt for details.
"""

from collections import deque
from copy import deepcopy
from datetime import datetime, timedelta
from json import dumps
from threading import Event
from typing import Deque, Optional, List, Any, Tuple

from PySide2.QtWidgets import (
    QLineEdit,
    QLabel,
    QHBoxLayout,
    QVBoxLayout,
    QWidget,
    QTabWidget,
    QSizePolicy,
    QTreeWidget,
)

from scine_database import Manager, Structure, ID
from scine_utilities import AtomCollection
from scine_chemoton.steering_wheel.datastructures import SelectionResult, ProtocolEntry
from scine_chemoton.steering_wheel.network_expansions import NetworkExpansion
from scine_chemoton.gears.elementary_steps import ElementaryStepGear
from scine_chemoton.gears.elementary_steps.trial_generator.fragment_based import FragmentBased
from scine_chemoton.gears.elementary_steps.trial_generator.bond_based import BondBased
from scine_chemoton.utilities.reactive_complexes.inter_reactive_complexes import assemble_reactive_complex
from scine_chemoton.utilities.datastructure_transfer import make_picklable

from scine_heron.chemoton.chemoton_widget import generate_gear_name_settings_suggestions
from scine_heron.chemoton.class_searcher import ChemotonClassSearcher
from scine_heron.chemoton.filter_builder import FilterBuilder
from scine_heron.chemoton.grouped_combo_box import GroupedComboBox
from scine_heron.settings.class_options_widget import generate_instance_based_on_potential_widget_input

from scine_heron.containers.progress_bar import HeronProgressBar
from scine_heron.containers.start_stop_widget import StartStopWidget
from scine_heron.containers.wrapped_label import WrappedLabel
from scine_heron.containers.buttons import TextPushButton
from scine_heron.containers.layouts import VerticalLayout
from scine_heron.utilities import (
    write_error_message,
    vertical_scroll_area_wrap,
    write_info_message,
    timedelta_string
)

from .display_widget import SteeringDisplay
from .structure_display import StructureWithReactiveSites
from .tree_items import TreeWidget, TreeWidgetItem
from .selection_threads import UnimolecularThread, BimolecularThread, UnimolecularCoordinates, BimolecularCoordinates


class CurrentSelectionDisplay(StartStopWidget):
    """
    The widget displaying the current selection.

    Notes
    -----
    It inherits from StartStopWidget, because we want to pass it to the ChemotonWidgetContainer,
    but it does not really implement any abstract methods, because these are not meant to be called
    for this widget.
    """

    def __init__(self, parent: SteeringDisplay, db_manager: Optional[Manager] = None,
                 selection: Optional[SelectionResult] = None, name: Optional[str] = None):
        super().__init__(parent=parent)
        label = QLabel("Display of current selection")
        self._layout.setContentsMargins(0, 0, 0, 0)
        self._layout.addWidget(label)
        if name is not None:
            name_label = WrappedLabel(name)
            self._layout.addWidget(name_label)

        self._selection = selection
        self._manager = db_manager
        if self._selection is None or self._manager is None:
            return
        self._selection.aggregate_filter.initialize_collections(self._manager)
        self._selection.reactive_site_filter.initialize_collections(self._manager)
        self._selection.further_exploration_filter.initialize_collections(self._manager)
        self._set_filter_info()
        self._content = vertical_scroll_area_wrap(SelectionContentWidget(self, self._manager, self._selection, parent))
        self._content.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Expanding)
        self._layout.addWidget(self._content)
        self.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Expanding)

    @classmethod
    def with_loading_info(cls, parent: SteeringDisplay):
        """
        Constructs a dummy widget that tells the user that we are currently constructing the widget.

        Notes
        -----
        Since the introduction of the update button, the load time is very short and this is not required
        any more. Can be removed when refactoring the update logic to avoid too many molecular widgets.
        (See SteeringDisplay.construct_new_current_selection_widget docs)

        Parameters
        ----------
        parent : SteeringDisplay
            The parent widget holding this widget

        Returns
        -------
        CurrentSelectionDisplay
            The dummy widget
        """
        inst = cls(parent=parent)
        query_info = QLineEdit("Loading current selection...")
        query_info.setReadOnly(True)
        inst.layout().addWidget(query_info)
        return inst

    def _set_filter_info(self):
        """
        Writes the filter information
        """
        filter_defaults = FilterBuilder.default_filters()
        for default, set_filter in zip(filter_defaults, [self._selection.aggregate_filter,
                                                         self._selection.reactive_site_filter,
                                                         self._selection.further_exploration_filter]):
            # ensure that comparison is made by explicit type equals and not 'isinstance' due to inheritance
            default_type = type(default)
            filter_type = type(default)
            if filter_type != default_type:
                self._add_filter_at_layout(set_filter.name)

    def join(self, force_join: bool = False) -> None:
        """
        Dummy abstract method overload, not meant to be called
        """

    def start_stop(self, start_all: bool = False, stop_all: bool = False) -> None:
        """
        Dummy abstract method overload, not meant to be called
        """

    def set_docstring_dict(self, doc_string_parser: Any) -> None:
        """
        Dummy abstract method overload, not meant to be called
        """

    def stop_class_if_working(self) -> None:
        """
        Dummy abstract method overload, not meant to be called
        """


class SelectionContentWidget(QWidget):
    """
    This widget displays and handles most of the current selection display.
    It is meant to be held by the CurrentSelectionWidget and wrapped by a vertical scrollbar.
    """

    def __init__(self, parent: CurrentSelectionDisplay, db_manager: Manager, selection_result: SelectionResult,
                 total_display: SteeringDisplay):
        """
        Construct the widget with parent, database, and selection information.
        The database is required because we query it to find out what possible network expansion would set up.

        Parameters
        ----------
        parent : CurrentSelectionDisplay
            The CurrentSelectionWidget that holds us
        db_manager : Manager
            The manager of the Scine Database
        selection_result : SelectionResult
            The selection result this widget represents
        total_display : SteeringDisplay
            The steering tab widget, currently only required for the model.
        """
        super().__init__(parent)
        self.setMaximumHeight(750)
        self._total_display = total_display
        # propagate db information
        self._manager = db_manager
        self._credentials = db_manager.get_credentials()
        self._structure_collection = db_manager.get_collection("structures")
        self._compound_collection = db_manager.get_collection("compounds")
        self._flask_collection = db_manager.get_collection("flasks")
        self._selection_result = selection_result
        self._selection_result.aggregate_filter.initialize_collections(db_manager)
        self._selection_result.reactive_site_filter.initialize_collections(db_manager)
        self._selection_result.further_exploration_filter.initialize_collections(db_manager)

        # set up tabs
        self._uni_tab = SelectionTab(self)
        self._bi_tab = SelectionTab(self)
        self._structure_tab = SelectionTab(self)
        self._uni_tab.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Expanding)
        self._bi_tab.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Expanding)
        self._structure_tab.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Expanding)
        self._tabs = QTabWidget(self)
        self._tabs.addTab(self._uni_tab, "Unimolecular")
        self._tabs.addTab(self._bi_tab, "Bimolecular")
        self._tabs.addTab(self._structure_tab, "Lone Structures")
        self._tabs.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Expanding)

        # get DB data
        self._lone_structures = self._gather_lone_structures()

        # set up potential next steps interface
        self._potential_next_step: Optional[NetworkExpansion] = None
        self._step_searcher = ChemotonClassSearcher(NetworkExpansion)
        self._chemical_step_box = GroupedComboBox(self, self._step_searcher)
        self._progress_bar = HeronProgressBar(self)
        self._site_settings_button = TextPushButton("Update", self.update_content, self)
        self._runtime_of_update = datetime.now()
        self._runtime_memory: Deque[timedelta] = deque([], maxlen=5)

        # update handling
        self._uni_update_thread: Optional[UnimolecularThread] = None
        self._bi_update_thread: Optional[BimolecularThread] = None
        self._finished_uni_update = Event()
        self._finished_bi_update = Event()
        self._uni_items: List[TreeWidgetItem] = []
        self._is_updating = False

        # put everything into a layout
        self._layout = VerticalLayout([
            self._chemical_step_box,
            self._progress_bar,
            self._site_settings_button,
            self._tabs
        ])
        self.setLayout(self._layout)
        self.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Expanding)

    def closeEvent(self, event):
        """
        Overload of the close event to ensure that the update threads are stopped.
        """
        self._cleanup_threads(active_stop=True)
        super().closeEvent(event)

    def _cleanup_threads(self, active_stop: bool = False) -> None:
        """
        Wait for update threads and join them.
        """
        # clear events for next run
        self._finished_uni_update.clear()
        self._finished_bi_update.clear()
        self._uni_items = []
        self._runtime_of_update = datetime.now()
        self._runtime_memory.clear()
        if self._uni_update_thread is not None:
            if active_stop:
                self._uni_update_thread.stop()
            self._uni_update_thread.wait()
            self._uni_update_thread = None
        if self._bi_update_thread is not None:
            if active_stop:
                self._bi_update_thread.stop()
            self._bi_update_thread.wait()
            self._bi_update_thread = None

    def _start_threads(self, potential_next_step: NetworkExpansion, selection: SelectionResult) -> None:
        """
        Starts the threads that gather the information about all uni- and bimolecular complexes set up with the
        given potential next exploration step.

        Parameters
        ----------
        potential_next_step : NetworkExpansion
            The potential next exploration step for which the complexes are gathered.
        selection : SelectionResult
            The selection result onto which the potential expansion should work on
        """
        self._cleanup_threads()

        # prepare transfer
        safe_step = make_picklable(potential_next_step)
        safe_selection = make_picklable(selection)
        if not isinstance(safe_step, NetworkExpansion):
            raise RuntimeError("Could not pickle potential next step")
        if not isinstance(safe_selection, SelectionResult):
            raise RuntimeError("Could not pickle selection result")
        assert deepcopy(safe_step)
        assert deepcopy(safe_selection)

        # construct threads
        self._uni_update_thread = UnimolecularThread(self, self._credentials, deepcopy(safe_step),
                                                     deepcopy(safe_selection))
        self._bi_update_thread = BimolecularThread(self, self._credentials, deepcopy(safe_step),
                                                   deepcopy(safe_selection))

        # signals that both have
        self._uni_update_thread.final_count_signal.connect(lambda count:  # pylint: disable=no-member
                                                           self._finish_update(0, count))
        self._uni_update_thread.info_message_signal.connect(write_info_message)  # pylint: disable=no-member
        self._uni_update_thread.error_message_signal.connect(write_error_message)  # pylint: disable=no-member
        self._uni_update_thread.loop_signal.connect(self._increase_progress_bar)  # pylint: disable=no-member

        self._bi_update_thread.final_count_signal.connect(lambda count:  # pylint: disable=no-member
                                                          self._finish_update(1, count))
        self._bi_update_thread.info_message_signal.connect(write_info_message)  # pylint: disable=no-member
        self._bi_update_thread.error_message_signal.connect(write_error_message)  # pylint: disable=no-member
        self._bi_update_thread.loop_signal.connect(self._increase_progress_bar)  # pylint: disable=no-member

        # uni signals
        self._uni_update_thread.single_structure_no_sites_signal.connect(  # pylint: disable=no-member
            self._add_uni_item
        )
        self._uni_update_thread.coordinates_signal.connect(self._fill_unimolecular)  # pylint: disable=no-member

        # bi signals
        self._bi_update_thread.coordinates_signal.connect(self._fill_bimolecular)  # pylint: disable=no-member

        # start
        self._uni_update_thread.start()
        self._bi_update_thread.start()

    def _add_uni_item(self, name: str, atoms: AtomCollection) -> None:
        """
        Builds a TreeWidgetItem with the given name and atoms and adds it to our container, which
        we add at once when we receive the finish signal.

        Parameters
        ----------
        name : str
            The name of the item.
        atoms : AtomCollection
            The atoms shown in the item
        """
        self._uni_items.append(TreeWidgetItem(self._uni_tab.tree, [name], atoms, []))

    def _fill_unimolecular(self, coordinates: UnimolecularCoordinates) -> None:
        total_count = 0
        for cid, sub_dict in coordinates.data.items():
            for sid, coords in sub_dict.items():
                if self._selection_result.structures and ID(sid) not in self._selection_result.structures:
                    continue
                total_count += self._single_structure_add(sid, self._uni_tab, cid, coords)
        self._finish_update(0, total_count)

    def _fill_bimolecular(self, coordinates_holder: BimolecularCoordinates) -> None:
        # we have to set up the protocol here, because the setup for the coordinates was done in a different process
        if self._potential_next_step is None:
            raise RuntimeError("No potential next expansion step set up")
        self._potential_next_step.dry_setup_protocol(self._credentials, self._selection_result)
        items = []
        total_count = 0
        for cid, sub_dict in coordinates_holder.data.items():
            compound_a, compound_b = cid.split("-")
            individual_indices: List[List[int]] = []
            individual_complexes: List[AtomCollection] = []
            individual_structure_pairs: List[List[str]] = []
            for sid, coords in sub_dict.items():
                structure_list = [Structure(ID(s), self._structure_collection) for s in sid.split("-")]
                if self._selection_result.structures and not all(ID(s) in self._selection_result.structures
                                                                 for s in sid.split("-")):
                    continue
                structure_a = structure_list[0]
                structure_b = structure_list[1]
                atoms_a = structure_a.get_atoms()
                atoms_b = structure_b.get_atoms()
                n_atoms_lhs = len(atoms_a)
                complex_index = 0
                for (coordinates, _), complexes in coords.items():
                    lhs: List[int] = []
                    rhs: List[int] = []
                    intra_total: List[int] = []
                    for pair in coordinates:
                        self._map_indices(lhs, rhs, intra_total, pair, n_atoms_lhs)
                    if complexes:
                        for complex in complexes:
                            complex_index = complex_index + 1
                            atoms, lhs, shifted_rhs = assemble_reactive_complex(atoms_a, atoms_b, lhs, rhs, *complex)
                            this_indices = lhs + shifted_rhs + intra_total
                            individual_indices.append(this_indices)
                            individual_complexes.append(atoms)
                            individual_structure_pairs.append([str(structure_a.id()), str(structure_b.id())])
                            assert all(s < len(atoms) for s in this_indices)
            if individual_indices and individual_complexes:
                total_indices = self.flatten_and_deduplicate_list(individual_indices)
                total_atoms = individual_complexes[0]
                top_item = TreeWidgetItem(self._bi_tab.tree, [f"Compounds ({compound_a}-{compound_b})"],
                                          total_atoms, total_indices)
                for i, (indices, complex_atoms, structure_pair) in \
                        enumerate(zip(individual_indices, individual_complexes, individual_structure_pairs)):
                    total_count += 1
                    TreeWidgetItem(top_item, ["Structures (\n\t" + '-\n\t'.join(s for s in structure_pair) +
                                              "\n\t_coordinate_" + str(i) + "\n)"],
                                   complex_atoms, indices)
                items.append(top_item)
        self._bi_tab.tree.insertTopLevelItems(0, items)
        self._finish_update(1, total_count)

    def _fill_structures(self) -> None:
        """
        Fills the tab with structures without an aggregate.
        Pretty fast and easy, because we gather the structures already in the constructor,
        because they don't change between different network expansion steps.
        """
        total_count = 0
        for sid in self._lone_structures:
            total_count += 1
            total_count += self._single_structure_add(sid, self._structure_tab)
        self._structure_tab.set_count(total_count)

    def _finish_update(self, index: int, count: int, active_stop: bool = False) -> None:
        """
        Handles the finishing of an update thread.
        """
        if index == 0:
            self._finished_uni_update.set()
            self._uni_tab.set_count(count)
            if self._uni_items:
                self._uni_tab.tree.insertTopLevelItems(0, self._uni_items)
                self._uni_items = []
                self._tabs.setCurrentIndex(0)
        elif index == 1:
            self._finished_bi_update.set()
            self._bi_tab.set_count(count)
            if count:
                self._tabs.setCurrentIndex(1)
        else:
            raise RuntimeError(f"Invalid index {index}")
        if self._finished_uni_update.is_set() and self._finished_bi_update.is_set():
            self._cleanup_threads(active_stop=active_stop)
            self._fill_progress_bar()
            self._site_settings_button.setText("Update")
            self._site_settings_button.setEnabled(True)
            self._is_updating = False
            self.updateGeometry()

    def update_content(self) -> None:
        """
        Update function that queries the database to build the reactive complexes and fills our tabs.
        This function may take some time.
        """
        if self._is_updating:
            write_info_message("Aborted preview")
            self._clear_and_terminate()
            return
        self._is_updating = True
        self._site_settings_button.setText("Abort")
        name = self._chemical_step_box.currentText().strip()

        model = self._total_display.get_model()
        predefined_kwargs = {
            "model": model,
        }
        # first create a dummy expansion from which we get the gears it will put in its protocol, so we can
        # give proper suggestions for the gear options
        dummy_step = self._step_searcher[name](**predefined_kwargs)
        dummy_step.dry_setup_protocol(self._credentials, self._selection_result)
        suggestions = generate_gear_name_settings_suggestions(dummy_step.current_gears())
        self._potential_next_step = \
            generate_instance_based_on_potential_widget_input(self, self._step_searcher[name], predefined_kwargs,
                                                              suggestions)
        dummy_step.protocol = []  # make sure we delete forked objects
        try:
            self._potential_next_step.dry_setup_protocol(self._credentials, self._selection_result)
            selection = {"exploration_disabled": {"$ne": True}}
            if hasattr(self._potential_next_step, "options") \
                    and getattr(self._potential_next_step.options, "react_flasks", False):
                self._progress_bar.setMaximum(max(2 * self._flask_collection.count(dumps(selection)), 1))
            else:
                self._progress_bar.setMaximum(max(2 * self._compound_collection.count(dumps(selection)), 1))
            self._progress_bar.setValue(0)
            self._progress_bar.setTextVisible(True)
            self._fill_tabs()
        except BaseException as e:
            import traceback
            traceback.print_exc()
            write_error_message(str(e))
            self._clear_and_terminate()

    def _clear_and_terminate(self) -> None:
        self._clear_tabs()
        self._progress_bar.setToolTip("Aborted")
        if not self._finished_uni_update.is_set():
            self._finished_uni_update.set()
            self._finish_update(0, 0, active_stop=True)
        if not self._finished_bi_update.is_set():
            self._finished_bi_update.set()
            self._finish_update(1, 0, active_stop=True)

    def _increase_progress_bar(self) -> None:
        current_value = self._progress_bar.value()
        max_value = self._progress_bar.maximum()
        self._progress_bar.setValue(current_value + 1)
        self._progress_bar.setToolTip(f"{int((current_value + 1) * 100 / max_value)} % finished")
        interval = 10
        if (current_value % interval) == 0:
            if current_value > (max_value / 2):
                self._runtime_memory.append(datetime.now() - self._runtime_of_update)
                if len(self._runtime_memory) == self._runtime_memory.maxlen:
                    # calculate remaining time
                    average_time = sum([t.total_seconds() for t in self._runtime_memory]) / self._runtime_memory.maxlen
                    remaining_time_steps = (max_value - current_value) / interval
                    remaining_time = average_time * remaining_time_steps
                    self._site_settings_button.setText(f"Abort (est. time left "
                                                       f"{timedelta_string(timedelta(seconds=remaining_time))})")
            self._runtime_of_update = datetime.now()

    def _fill_progress_bar(self) -> None:
        self._progress_bar.setValue(self._progress_bar.maximum())

    def _fill_tabs(self) -> None:
        """
        The main updating function that takes the most time, but the two time intensive tabs are threaded away.
        """
        self._clear_tabs()
        if self._potential_next_step is None:
            raise RuntimeError("No potential next expansion step set up")
        self._potential_next_step.protocol.clear()  # we clear here, because we have to transfer across
        self._start_threads(self._potential_next_step, self._selection_result)
        self._fill_structures()

    def _clear_tabs(self) -> None:
        """
        Clears all sub tabs
        """
        self._uni_tab.clear()
        self._bi_tab.clear()
        self._structure_tab.clear()

    def _map_indices(self, lhs_to_add_to: List[int], rhs_to_add_to: List[int], intra_to_add_to: List[int],
                     pair: Tuple[int, int], n_atoms_lhs: int) -> None:
        """
        This is a utility method that ensures indexing consistency that is required,
        because Chemoton has a varying API logic for its indices for bimolecular cases.

        Parameters
        ----------
        lhs_to_add_to : List[int]
            The list of left-hand-side indices we add indices to
        rhs_to_add_to : List[int]
            The list of right-hand-side indices we add indices to
        intra_to_add_to : List[int]
            The list of intramolecular indices we add indices to
        pair : Tuple[int, int]
            The reactive indices
        n_atoms_lhs : int
            The number of atoms of the left-hand-side structure

        Raises
        ------
        NotImplementedError
            Multiple elementary step gears with different TrialGenerators
        NotImplementedError
            An unknown TrialGenerator
        RuntimeError
            No gear with a TrialGenerator present
        """
        # find out TrialGenerator
        trial_generator = None
        assert self._potential_next_step is not None
        for entry in self._potential_next_step.protocol:
            if not isinstance(entry, ProtocolEntry):
                continue
            if isinstance(entry.gear, ElementaryStepGear):
                if entry.gear.options.enable_bimolecular_trials:
                    if trial_generator is None:
                        trial_generator = entry.gear.trial_generator
                    elif not issubclass(type(trial_generator), type(entry.gear.trial_generator)) \
                            and not issubclass(type(entry.gear.trial_generator), type(trial_generator)):
                        raise NotImplementedError("Varying Trial Generator currently not supported")
        if trial_generator is None:
            raise RuntimeError("Could not determine Trial Generator")

        # fill in lists depending on TrialGenerator
        if isinstance(trial_generator, FragmentBased):
            assert not any(p >= n_atoms_lhs for p in pair)
            lhs_to_add_to.append(pair[0])
            rhs_to_add_to.append(pair[1])
        elif isinstance(trial_generator, BondBased):
            p1 = pair[0]
            p2 = pair[1]
            if p1 >= n_atoms_lhs > p2:
                lhs_to_add_to.append(p2)
                rhs_to_add_to.append(p1 - n_atoms_lhs)
            elif p2 >= n_atoms_lhs > p1:
                lhs_to_add_to.append(p1)
                rhs_to_add_to.append(p2 - n_atoms_lhs)
            else:
                intra_to_add_to.append(p1)
                intra_to_add_to.append(p2)
        else:
            raise NotImplementedError(f"Trial Generator {type(trial_generator)} currently not supported")

    def _single_structure_add(self, structure_id: str, tab_to_add_to: QTreeWidget, compound_id: Optional[str] = None,
                              coordinates: Optional[List[Tuple[List[List[Tuple[int, int]]], int]]] = None) -> int:
        """
        Adds a single structure to a tab.

        Parameters
        ----------
        structure_id : str
            The ID of the structure as a string
        tab_to_add_to : TreeWidget
            The tab we want to add to, type hint in code is not explicit due to circular dependency
        compound_id : Optional[str], optional
            The compound ID the structure belongs to as a string, by default None
        coordinates : Optional[List[Tuple[List[List[Tuple[int, int]]], int]]], optional
            Reactive coordinates already cleaned by index mapping, by default None.
            If not given, only a top level item is added and no reactive complexes with reactive indices are shown.

        Returns
        -------
        int
            The number of structures (i.e. individual subitems)
        """
        total_count = 0
        structure = Structure(ID(structure_id), self._structure_collection)
        if coordinates is None or not structure.has_graph("masm_cbor_graph") \
                or not structure.has_graph("masm_idx_map"):
            top_item = TreeWidgetItem(tab_to_add_to.tree, [f"Structure ({structure_id})"], structure.get_atoms(), [])
        else:
            individual_indices: List[List[int]] = []
            for coords, _ in coordinates:
                for pairs in coords:
                    this_indices: List[int] = []
                    for pair in pairs:
                        this_indices.append(pair[0])
                        this_indices.append(pair[1])
                    individual_indices.append(this_indices)
            total_indices = self.flatten_and_deduplicate_list(individual_indices)
            if not total_indices:
                # no reaction possible with this, so we don't show it
                return 0
            atoms = structure.get_atoms()
            top_item = TreeWidgetItem(tab_to_add_to.tree, [f"Compound ({compound_id})"],
                                      atoms, total_indices)
            for count, indices in enumerate(individual_indices):
                total_count += 1
                TreeWidgetItem(top_item, [f"Structure ({structure_id}_coordinate_{count}"],
                               atoms, indices)
        tab_to_add_to.tree.insertTopLevelItems(0, [top_item])
        return total_count

    def _gather_lone_structures(self) -> List[str]:
        """
        Identifies all the individual structures in the given selection result, that don't have an aggregate.

        Returns
        -------
        List[str]
            The IDs of the structures as strings
        """
        lone_structures: List[str] = []
        if self._selection_result.structures:
            for sid in self._selection_result.structures:
                structure = Structure(sid, self._structure_collection)
                if not structure.has_aggregate():
                    lone_structures.append(str(sid))
        return lone_structures

    @staticmethod
    def flatten_and_deduplicate_list(list_of_lists: List[List[int]]) -> List[int]:
        """
        Utility function to flatten list of lists and deduplicate the resulting list.
        """
        return list(set([index for sublist in list_of_lists for index in sublist]))


class SelectionTab(QWidget):
    """
    The widget that represents a tab and contains the selection tree and the molecule viewer.
    """

    def __init__(self, parent: Optional[QWidget]) -> None:
        super().__init__(parent)
        count_layout = QHBoxLayout()
        count_label = QLabel("Total job count:")
        self._count_edit = QLineEdit("0")
        self._count_edit.setReadOnly(True)
        count_layout.addWidget(count_label)
        count_layout.addWidget(self._count_edit)
        self.tree = TreeWidget(parent=self)
        self.tree.setColumnCount(1)
        self.tree.setHeaderHidden(True)
        self._mol_widget = StructureWithReactiveSites(parent=self)

        self._layout = QVBoxLayout()
        self._layout.addLayout(count_layout)
        self._layout.addWidget(self.tree)
        self._layout.addWidget(self._mol_widget)
        self.setLayout(self._layout)

    def set_count(self, count: int) -> None:
        """
        Set the count display
        """
        self._count_edit.setText(str(count))

    def display_atoms_with_sites(self, atoms: AtomCollection, indices: List[int]) -> None:
        """
        Plug in new structure and reactive sites.

        Parameters
        ----------
        atoms : AtomCollection
            The new structure
        indices : List[int]
            The new reactive indices
        """
        self._mol_widget.update_structure(atoms, indices)

    def reset_camera(self):
        """
        Reset the camera in the molecular viewer
        """
        self._mol_widget.reset_camera()

    def clear(self) -> None:
        """
        Clear everything the tab is holding
        """
        self.tree.clear()
        self._count_edit.setText("0")
        self._mol_widget.clear()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
__copyright__ = """ This code is licensed under the 3-clause BSD license.
Copyright ETH Zurich, Department of Chemistry and Applied Biosciences, Reiher Group.
See LICENSE.txt for details.
"""

from typing import Optional
from PySide2.QtCore import QObject
from PySide2.QtWidgets import (
    QLineEdit,
    QInputDialog,
    QDialog,
    QLabel,
    QVBoxLayout,
    QWidget
)


def text_input_box(parent: Optional[QWidget], label: str, question: str, input_field: QLineEdit, hint: str) -> str:
    """
    Create a pop-up window with a single text input field and return text, when the window is closed.

    Parameters
    ----------
    parent : Optional[QWidget]
        The parent of the pop-up window
    label : str
        The label of the pop-up window
    question : str
        The question to be asked before the input field
    input_field : QLineEdit
        The input field
    hint : str
        A prefilled text in the input field

    Returns
    -------
    str
        The entered text
    """
    dialog = QInputDialog(parent=parent)
    return dialog.getText(dialog.parent(), label, question, input_field, hint)[0].strip()


def yes_or_no_question(parent: QObject, question: str, default_answer: str = "yes") -> bool:
    """
    Ask a yes or no question in a pop-up window in which the user can enter
    'y', 'yes', 'n' or 'no' and return True or False, respectively.

    Parameters
    ----------
    parent : QObject
        The parent of the pop-up window
    question : str
        The question to be asked

    Returns
    -------
    bool
        The answer to the question as a bool
    """
    answer = text_input_box(parent, "Scine Heron", f"{question} (y/n)? ", QLineEdit.Normal, default_answer)
    while True:
        if answer.lower().strip() in ["y", "yes"]:
            return True
        if answer.lower().strip() in ["n", "no"]:
            return False
        answer = text_input_box(parent, "Scine Heron", "Did not understand your answer, please write 'yes' or 'no'",
                                QLineEdit.Normal, "yes")


def pop_up_message(parent: QObject, message: str) -> None:
    """
    A pop-up window with a message.

    Parameters
    ----------
    parent : QObject
        The parent of the pop-up window
    message : str
        The message to be displayed
    """
    dialog = QDialog(parent)
    dialog.setWindowTitle("Scine Heron Message")
    layout = QVBoxLayout()
    layout.addWidget(QLabel(message))
    dialog.setLayout(layout)
    dialog.exec_()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
__copyright__ = """ This code is licensed under the 3-clause BSD license.
Copyright ETH Zurich, Department of Chemistry and Applied Biosciences, Reiher Group.
See LICENSE.txt for details.
"""
from os import path
from typing import List, Optional, TYPE_CHECKING
import pickle

from PySide2.QtWidgets import (
    QWidget,
    QInputDialog,
)
from PySide2.QtGui import QCloseEvent

import scine_utilities as su

from scine_heron.calculators.hybrid_model_construction import MMParametrizer, QMRegionSelector, SwooseTask
from scine_heron.containers.layouts import VerticalLayout
from scine_heron.containers.buttons import TextPushButton
from scine_heron.io.text_box import yes_or_no_question, pop_up_message
from scine_heron.settings.class_options_widget import ClassOptionsWidget
from scine_heron.settings.settings_status_manager import SettingsStatusManager
from scine_heron.utilities import write_error_message, write_info_message, docstring_dict_from_scine_settings
from scine_heron.dependencies.optional_import import is_imported

if TYPE_CHECKING:
    from scine_heron.database.connection_dialog import DatabaseConnectionDialog
    from scine_database import Manager
else:
    from scine_heron.dependencies.optional_import import importer
    DatabaseConnectionDialog = importer("scine_heron.database.connection_dialog", "DatabaseConnectionDialog")
    Manager = importer("scine_database", "Manager")


class SwooseSettingsWidget(QWidget):

    def __init__(self, parent: Optional[QWidget], settings_status_manager: SettingsStatusManager) -> None:
        super().__init__(parent)
        self.__settings_status_manager = settings_status_manager
        self.__swoose_backup_file = ".swoose_qm_region_backup.pkl"

        # threads
        self.__swoose_parametrize_task: Optional[SwooseTask] = None
        self.__swoose_select_task: Optional[SwooseTask] = None
        self.__setup_param_task()
        self.__setup_qm_selection_task()

        # buttons
        self.__mm_parametrize_button = TextPushButton("Parametrize SFAM", self.__parametrize_sfam)
        self.__mm_parametrize_button.setVisible(False)
        self.__qmmm_selection_button = TextPushButton("Construct QM region", self.__construct_hybrid_model)
        self.__qmmm_selection_button.setVisible(False)
        self.__qmmm_direct_selection_button = TextPushButton("Current selection as QM region",
                                                             self.__specify_current_selection_as_qm_region)
        self.__qmmm_direct_selection_button.setVisible(False)

        # connect settings status manager
        self.__settings_status_manager.hamiltonian_changed.connect(
            self.__determine_hybrid_model_selection_visibility
        )
        self.__settings_status_manager.hamiltonian_changed.connect(
            self.__determine_parametrize_option_visibility
        )

        # layout buttons
        layout = VerticalLayout([
            self.__mm_parametrize_button,
            self.__qmmm_selection_button,
            self.__qmmm_direct_selection_button
        ])
        self.setLayout(layout)

    def closeEvent(self, event: QCloseEvent) -> None:
        if self.__swoose_parametrize_task is not None and self.__swoose_parametrize_task.isRunning():
            self.__swoose_parametrize_task.terminate()
            self.__swoose_parametrize_task.wait()
        if self.__swoose_select_task is not None and self.__swoose_select_task.isRunning():
            self.__swoose_select_task.terminate()
            self.__swoose_select_task.wait()
        super().closeEvent(event)

    def __setup_param_task(self) -> None:
        if self.__swoose_parametrize_task is not None:
            if self.__swoose_parametrize_task.isRunning():
                self.__swoose_parametrize_task.terminate()
            self.__swoose_parametrize_task.wait()
        self.__swoose_parametrize_task = SwooseTask(self, MMParametrizer)
        # connect error handling
        self.__swoose_parametrize_task.error_signal.connect(
            write_error_message
        )
        # connect settings handling
        self.__swoose_parametrize_task.send_settings_signal.connect(
            self.__handle_swoose_settings
        )
        # connect results
        self.__swoose_parametrize_task.parametrize_signal.connect(
            self.__handle_parametrize_result
        )

    def __setup_qm_selection_task(self) -> None:
        if self.__swoose_select_task is not None:
            if self.__swoose_select_task.isRunning():
                self.__swoose_select_task.terminate()
            self.__swoose_select_task.wait()
        self.__swoose_select_task = SwooseTask(self, QMRegionSelector)
        # connect error handling
        self.__swoose_select_task.error_signal.connect(
            write_error_message
        )
        # connect settings handling
        self.__swoose_select_task.send_settings_signal.connect(
            self.__handle_swoose_settings
        )
        # connect results
        self.__swoose_select_task.qm_region_signal.connect(
            self.__handle_selection_result
        )

    def __determine_parametrize_option_visibility(self) -> None:
        method_family = self.__settings_status_manager.get_calculator_args()[0]
        mm_methods = ["SFAM", "GAFF"]
        self.__mm_parametrize_button.setVisible(method_family.upper() in mm_methods or '/' in method_family)

    def __determine_hybrid_model_selection_visibility(self) -> None:
        method_family = self.__settings_status_manager.get_calculator_args()[0]
        have_hybrid_model: bool = '/' in method_family
        self.__qmmm_selection_button.setVisible(have_hybrid_model)
        self.__qmmm_direct_selection_button.setVisible(have_hybrid_model)

    def __handle_swoose_settings(self, settings: su.Settings, identifier: str) -> None:
        method_family, program = self.__settings_status_manager.get_calculator_args()
        if su.settings_names.method_family in settings:
            settings[su.settings_names.method_family] = method_family
        if su.settings_names.program in settings:
            settings[su.settings_names.program] = program
        doc_strings = docstring_dict_from_scine_settings(settings)
        widget = ClassOptionsWidget(
            options=settings,
            docstring=doc_strings,
            parent=self,
            allow_removal=False
        )
        widget.exec_()
        for task in [self.__swoose_parametrize_task, self.__swoose_select_task]:
            if task is not None and task.identifier == identifier:
                task.receive_settings_signal.emit(settings)

    def __set_swoose_buttons_enabled(self, value: bool) -> None:
        for button in [self.__mm_parametrize_button, self.__qmmm_selection_button, self.__qmmm_direct_selection_button]:
            button.setEnabled(value)

    def __wrap_db_usage_around_task_start(self, task: SwooseTask, nomen: str, verb: str) -> None:
        self.__set_swoose_buttons_enabled(False)
        if not is_imported(Manager) or not is_imported(DatabaseConnectionDialog):
            # no database available, start direct mode
            task.disable_database()
            task.start()
            return
        db_direct_decision = QInputDialog()
        db_direct_decision.setWindowTitle("Mode decision")
        mode = db_direct_decision.getItem(self,
                                          nomen,
                                          f"Do you want to {verb} directly or via a database?",
                                          ["direct", "database"],
                                          current=0,
                                          editable=False)[0]
        if mode == "database":
            from scine_heron import find_main_window
            main = find_main_window()
            manager = None
            if main is not None:
                credentials = main.toolbar.current_credentials()
                if credentials is not None:
                    manager = Manager()
                    manager.set_credentials(credentials)
            while True:
                connect_dialog = DatabaseConnectionDialog(self, db_manager=manager)
                manager = connect_dialog.get_db_manager()
                if manager is not None and manager.connected:
                    task.enable_database(manager)
                    break
                else:
                    really_no_database = yes_or_no_question(self,
                                                            "Not connected to database. "
                                                            "Do you want to use the direct mode")
                    if really_no_database:
                        break
                    skip = yes_or_no_question(self, f"Do you want to abort the {nomen}")
                    if skip:
                        self.__set_swoose_buttons_enabled(True)
                        return
        else:
            task.disable_database()
        task.start()

    def __parametrize_sfam(self) -> None:
        if self.__swoose_parametrize_task is None:
            self.__setup_param_task()
            assert self.__swoose_parametrize_task is not None
        self.__wrap_db_usage_around_task_start(self.__swoose_parametrize_task, "Parametrization", "parametrize")

    def __construct_hybrid_model(self) -> None:
        if path.exists(self.__swoose_backup_file):
            ans = yes_or_no_question(self, f"Found file '{self.__swoose_backup_file}'. Do you want to load it instead?")
            if ans:
                sele = self.__load_qm_selection()
                self.__handle_selection_result(sele)
                return
            else:
                write_info_message("Ignoring backup file, starting selection from scratch")
        if self.__swoose_select_task is None:
            self.__setup_qm_selection_task()
            assert self.__swoose_select_task is not None
        self.__wrap_db_usage_around_task_start(self.__swoose_select_task, "QM region selection", "select QM region")

    def __handle_parametrize_result(self, result: bool) -> None:
        if result:
            pop_up_message(self, "Finished parametrization")
        else:
            write_error_message("Failed to parametrize structure")
        self.__setup_param_task()
        self.__set_swoose_buttons_enabled(True)

    def __handle_selection_result(self, result: List[int]) -> None:
        from scine_heron.molecular_viewer import get_mol_viewer_tab
        self.__set_swoose_buttons_enabled(True)
        if not result:
            return
        store_message = f"Storing result in '{self.__swoose_backup_file}' in case you want to enter it manually"
        mol_tab = get_mol_viewer_tab(want_atoms_there=True)
        if mol_tab is None:
            write_error_message(f"Got QM region selection, but could not access structure in Molecular Viewer. "
                                f"{store_message}")
            self.__save_qm_selection(result)
            return
        result_string = "["
        split_size = 10
        for i in range(0, str(result).count(',') + 1, split_size):
            result_string += ",".join([str(r) for r in result[i:i + split_size]]) + ",\n"
        result_string += "]"
        answer = yes_or_no_question(self, f"Determined QM region {result}. Do you want to set this in the calculator?")
        if not answer:
            write_info_message(store_message)
            self.__save_qm_selection(result)
            return
        assert mol_tab.mol_widget is not None
        n_atoms = len(mol_tab.mol_widget.get_atom_collection())
        if any(r >= n_atoms for r in result):
            write_error_message(f"Got QM region selection, but it does not fit in the current structure. "
                                f"{store_message}")
            self.__save_qm_selection(result)
            return
        mol_tab.mol_widget.set_selection(result)
        self.__specify_current_selection_as_qm_region()
        self.__setup_qm_selection_task()

    def __save_qm_selection(self, selection: List[int]) -> None:
        with open(self.__swoose_backup_file, "wb") as f:
            pickle.dump(selection, f)

    def __load_qm_selection(self) -> List[int]:
        with open(self.__swoose_backup_file, "rb") as f:
            return pickle.load(f)

    def __specify_current_selection_as_qm_region(self) -> None:
        from scine_heron.molecular_viewer import get_mol_viewer_tab
        mol_tab = get_mol_viewer_tab(want_atoms_there=True)
        if mol_tab is None:
            write_error_message("Could not get selection, no structure present")
            return
        mol_widget = mol_tab.mol_widget
        assert mol_widget is not None
        selection = mol_widget.get_selection()
        if not selection:
            write_error_message("Could not get selection, nothing selected")
            return
        method_family = self.__settings_status_manager.get_calculator_args()[0]
        have_hybrid_model: bool = '/' in method_family
        if not have_hybrid_model:
            write_error_message("No hybrid method family such as 'PM6/SFAM' currently loaded")
            return
        # TODO move Swoose settingsname into settingsnames in utils and use binding instead of naked string
        mol_tab.create_calculator_widget.update_settings({"qm_atoms": selection})
        self.__set_swoose_buttons_enabled(True)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
__copyright__ = """ This code is licensed under the 3-clause BSD license.
Copyright ETH Zurich, Department of Chemistry and Applied Biosciences, Reiher Group.
See LICENSE.txt for details.
"""
"""
Provides the MoleculeWidget class.
"""

import multiprocessing
from multiprocessing.managers import ListProxy
from uuid import uuid1
from pathlib import Path
from typing import Optional, Tuple, List, Any, Dict, Union, TYPE_CHECKING
if TYPE_CHECKING:
    Signal = Any
else:
    from PySide2.QtCore import Signal

import scine_utilities as utils
import numpy as np
from vtkmodules.vtkCommonExecutionModel import vtkTrivialProducer
from vtk import (
    vtkActor,
    vtkMolecule,
    vtkMoleculeMapper,
    vtkRenderer,
    vtkSelectVisiblePoints,
    vtkSimpleBondPerceiver,
    vtkDoubleArray,
    vtkUnsignedCharArray,
    vtkDataObject,
    vtkArrowSource,
    vtkPolyDataMapper,
    vtkTransformPolyDataFilter,
    vtkTransform,
)
from vtkmodules.qt.QVTKRenderWindowInteractor import QVTKRenderWindowInteractor
from PySide2.QtCore import QObject, Qt
from PySide2.QtWidgets import QHBoxLayout, QWidget
from PySide2.QtGui import QGuiApplication

import scine_heron.config as config
from scine_heron.calculators.calculator import ScineCalculatorWrapper
from scine_heron.molecule.depth_view import DepthViewZ1dFixedOnHapticSoftMaxMin
from scine_heron.edit_molecule import edit_molecule_functions as emf
from scine_heron.status_manager import StatusManager
from scine_heron.haptic.haptic_client import HapticClient
from scine_heron.molecule.colorbar import (
    create_colorbar,
    create_symmetric_color_transfer_function,
)
from scine_heron.molecule.utils.molecule_utils import (
    maximum_vdw_radius,
    atom_collection_to_molecule,
    molecule_to_atom_collection,
)
from scine_heron.molecule.utils.array_utils import rescale_to_range, iterable_to_vtk_array
from scine_heron.molecule.atom_selection import AtomSelection
from scine_heron.molecule.molecule_labels import MoleculeLabels
from scine_heron.electronic_data.electronic_data_widget import ElectronicDataWidget
from scine_heron.energy_profile.energy_profile_status_manager import (
    EnergyProfileStatusManager,
)
from scine_heron.settings.settings_status_manager import SettingsStatusManager
from scine_heron.settings.settings import MoleculeStyle
from scine_heron.molecule.styles.molecule_interactor_style import MoleculeInteractorStyle
from scine_heron.mediator_potential.custom_results import CustomResult
from scine_heron.utilities import write_error_message, hex_to_rgb_base_1


class WrappedQVTKRenderWindowInteractor(QVTKRenderWindowInteractor):
    def __init__(self, *args, **kwargs):
        if 'alternative_zoom_controls' in kwargs:
            self.__alternative_zoom_controls = kwargs['alternative_zoom_controls']
            del kwargs['alternative_zoom_controls']
        else:
            self.__alternative_zoom_controls = False
        if 'disable_modification' in kwargs:
            self.__disable_modification = kwargs['disable_modification']
            del kwargs['disable_modification']
        else:
            self.__disable_modification = False
        QVTKRenderWindowInteractor.__init__(self, *args, **kwargs)

    def wheelEvent(self, ev):
        if self.__alternative_zoom_controls:
            modifiers = QGuiApplication.keyboardModifiers()
            if modifiers == Qt.ControlModifier:
                super().wheelEvent(ev)
            else:
                pass
        else:
            super().wheelEvent(ev)

    def mousePressEvent(self, ev):
        if self.__disable_modification and ev.button() == Qt.RightButton:
            pass
        else:
            super().mousePressEvent(ev)

    def mouseReleaseEvent(self, ev):
        if self.__disable_modification and ev.button() == Qt.RightButton:
            pass
        else:
            super().mouseReleaseEvent(ev)


class MoleculeWidget(QWidget):
    """
    Displays molecule in a 3D view.
    """

    total_number_of_mol_viewers = 0
    settings_changed_signal = Signal(int, int, str)

    def __init__(
        self,
        settings_status_manager: SettingsStatusManager = SettingsStatusManager(),
        parent: Optional[QObject] = None,
        file: Optional[Path] = None,
        atoms: Optional[object] = None,
        bonds: Optional[object] = None,
        haptic_client: Optional[HapticClient] = None,
        energy_status_manager: Optional[EnergyProfileStatusManager] = None,
        alternative_zoom_controls: bool = False,
        disable_modification: bool = False,
    ):
        QWidget.__init__(self, parent)
        if MoleculeWidget.total_number_of_mol_viewers < 100:
            MoleculeWidget.total_number_of_mol_viewers += 1
        else:
            msg = "Cannot open more MoleculeWidgets only 100 are allowed."
            write_error_message(msg)
            raise RuntimeError(msg)

        self.__disable_modification: bool = disable_modification
        self.__molecule_version = uuid1()
        self.__settings_status_manager = settings_status_manager
        self.__settings_status_manager.molecule_style_changed.connect(
            self.__update_molecule_style_if_necessary)
        self.__settings_status_manager.labels_style_changed.connect(
            self.__update_labels_style)
        self.__energy_status_manager = energy_status_manager
        self.__charge_status_manager = StatusManager[Optional[List[float]]](None)
        self.__force_status_manager = StatusManager[Optional[List[float]]](None)

        for signal in [
            self.__charge_status_manager.changed_signal,
            self.__settings_status_manager.molecule_style_changed,
        ]:
            signal.connect(self.__update_atomic_charge_display)
        self.__force_status_manager.changed_signal.connect(self.__update_force_arrow_display)
        self.__colorbar = create_colorbar()
        self.__haptic_client = haptic_client

        # setup
        self.__filter = vtkSimpleBondPerceiver()
        # We do not SetInputData(self.__molecule) just yet
        self.__molecule_mapper = vtkMoleculeMapper()
        self.__molecule_mapper.SetInputConnection(self.__filter.GetOutputPort())

        self.__molecule_actor = vtkActor()
        self.__molecule_actor.SetMapper(self.__molecule_mapper)

        self.__renderer = vtkRenderer()

        self.__interactor = WrappedQVTKRenderWindowInteractor(
            parent=self,
            alternative_zoom_controls=alternative_zoom_controls,
            disable_modification=disable_modification
        )
        self.__interactor.Initialize()
        self.__interactor.GetRenderWindow().AddRenderer(self.__renderer)

        self.__renderer.AddActor(self.__molecule_actor)

        self.__selector = self.__create_selector(self.__renderer)
        self.__labels = MoleculeLabels(
            style=self.__settings_status_manager.labels_style, selector=self.__selector
        )
        self.__renderer.AddActor2D(self.__labels.actor)

        self.__actors = {
            "molecule": self.__molecule_actor,
            "labels": self.__labels.actor,
        }
        # normal selection
        self._atom_selection = AtomSelection()
        self.__renderer.AddActor(self._atom_selection.actor)
        self.__actors['atom_selection'] = self._atom_selection.actor

        # qm selection for qm/mm
        self._qm_selection = AtomSelection()
        self._qm_selection.change_color(*list(np.array([118, 199, 115]) / 256.0))
        self.__renderer.AddActor(self._qm_selection.actor)
        self.__actors['qm_selection'] = self._qm_selection.actor

        self.__electronic_data_widget = ElectronicDataWidget(
            self.__renderer, self.__settings_status_manager, self
        )
        self.__settings_status_manager.hamiltonian_changed.connect(  # pylint: disable=no-member
            self.__electronic_data_widget.clear_surfaces_and_data
        )

        self.__renderer.SetPreserveColorBuffer(True)
        self.__renderer.SetLayer(0)
        rgb = hex_to_rgb_base_1(config.COLORS['secondaryLightColor'])
        self.__renderer.SetBackground(*rgb)

        self.__style = MoleculeInteractorStyle(
            self.__interactor,
            self.__renderer,
            self.__molecule_mapper,
            self.__haptic_client,
            self.__actors,
            self.append_selection,
        )
        self.__style.set_status_managers(
            self.__settings_status_manager,
            self.__energy_status_manager,
            self.__charge_status_manager,
            self.__force_status_manager,
            self.__electronic_data_widget.electronic_data_status_manager,
        )
        self.__interactor.SetInteractorStyle(self.__style)
        self.__selected_atoms: List[int] = []
        self.__arrow_filter_list: List[vtkTransformPolyDataFilter] = []

        """ Add a 3D viewer to the widget """
        self.__depth_view = None
        if self.__must_show_depth_view():
            """ setup_depth_view """
            self.__depth_view = DepthViewZ1dFixedOnHapticSoftMaxMin(parent=self)

            self.__depth_molecule_producer = vtkTrivialProducer()
            self.__depth_view.set_molecule_input(
                self.__depth_molecule_producer.GetOutputPort()
            )

            self.__depth_view.set_haptic_pointer_data_input(
                self.__style.haptic_pointer.output
            )
            self.__renderer.GetRenderWindow().AddObserver(
                "RenderEvent", self.__depth_view.camera_changed
            )
            """
            Adds a zero margin layout with the viewer,
            including a depth view.
            Depending on the type of the view,
            a different layout is chosen.
            """
            layout = QHBoxLayout(self)
            layout.setMargin(0)
            layout.addWidget(self.__interactor, stretch=4)
            layout.addWidget(self.__depth_view.interactor, stretch=1)
            self.setLayout(layout)
        else:
            self.__add_layout()

        self.__loaded_molecule_once = False
        if file is not None:
            self.__load_from_file(file)
            if self.__haptic_client is not None:
                self.__haptic_client.update_molecule(self.__molecule)
                self.__settings_status_manager.number_of_molecular_orbital = None
            self.__loaded_molecule_once = True
        elif atoms is not None:
            self.__load_from_atom_collection(atoms, bonds)
            self.__update_haptic_client()
            self.__loaded_molecule_once = True
        else:
            self.__plug_in_new_molecule(vtkMolecule())
        self.__renderer.ResetCamera()

    def get_electronic_data_widget(self):
        return self.__electronic_data_widget

    def update_molecule(
        self,
        file: Optional[Path] = None,
        atoms: Optional[object] = None,
        bonds: Optional[object] = None,
        reset_camera: bool = True
    ):
        if self.__electronic_data_widget is not None:
            self.__electronic_data_widget.clear_surfaces_and_data()
        if file is not None:
            self.__load_from_file(file)
            if self.__haptic_client is not None:
                self.__haptic_client.update_molecule(self.__molecule)
                self.__settings_status_manager.number_of_molecular_orbital = None
        if atoms is not None:
            self.__load_from_atom_collection(atoms, bonds)
            self.__update_haptic_client()
        if not self.__loaded_molecule_once:
            self.__loaded_molecule_once = True
            self.__renderer.ResetCamera()
        elif reset_camera:
            self.__renderer.ResetCamera()
        if not self.__disable_modification:
            self.__create_force_arrows()

    def reset_camera(self):
        self.__renderer.ResetCamera()

    def __must_show_depth_view(self) -> bool:
        return (
            self.__haptic_client is not None
            and self.__haptic_client.device_is_available
        )

    def __update_haptic_client(self) -> None:
        if self.__haptic_client is not None:
            self.__haptic_client.update_molecule(self.__molecule)
            self.__settings_status_manager.selected_molecular_orbital = None
            self.__settings_status_manager.number_of_molecular_orbital = None

    def __update_atomic_charge_display(self) -> None:
        """
        Updates the display of the population analysis/atomic charge.

        If the display is disabled, or the values are invalid,
        (this happens if e.g. no Sparrow computation has completed),
        then the usual colors and style-defined labels are shown.
        """
        enabled = (
            self.__settings_status_manager.molecule_style
            == MoleculeStyle.PartialCharges
        )
        values = self.__charge_status_manager.value
        if (
            enabled
            and values is not None
            and len(values) == self.__molecule.GetNumberOfAtoms()
        ):
            array = iterable_to_vtk_array((round(x, 2) for x in values), len(values))
            self.display_array_via_colors(array)
            self.display_array_via_labels(array)
            radii = iterable_to_vtk_array((-1.0 * x for x in values), len(values))
            self.display_array_via_radii(radii)
        else:
            self.display_atomic_number_colors()
            self.display_style_labels()
            self.display_atomic_number_radii()
        self.__render_molecule()

    def __update_force_arrow_display(self) -> None:
        """
        Updates gradient arrows for all atoms in system.
        """
        enabled = (
            self.__settings_status_manager.molecule_style
            == MoleculeStyle.ForceArrow
        )
        values = self.__force_status_manager.value
        n_atoms = self.__molecule.GetNumberOfAtoms()
        for id in range(n_atoms):
            transform = self.__arrow_filter_list[id].GetTransform()
            if (
                enabled
                and values is not None
                and len(self.__arrow_filter_list) == n_atoms
            ):
                gradient = values[id]
                force2 = np.dot(gradient, gradient)
                if force2 > 10E-4 and force2 < 10:
                    scale = -40
                    # transform.Identity will set the arrow to (1, 0, 0)
                    transform.Identity()
                    force = np.sqrt(force2)
                    unit_gradient = gradient / force
                    atom = self.__molecule.GetAtom(id)
                    position = atom.GetPosition()
                    # Translation has to be done before rotation!
                    transform.Translate(position)
                    # Rotate identity arrow (1,0,0) to direction of gradient
                    transform.RotateWXYZ(
                        np.degrees(np.arccos(unit_gradient[0])),
                        np.cross([1, 0, 0], unit_gradient)
                    )
                    # Scale by force
                    transform.Scale(scale * force, scale * force, scale * force)
                else:
                    transform.Scale(0.0, 0.0, 0.0)
            else:
                transform.Scale(0.0, 0.0, 0.0)
            self.__arrow_filter_list[id].Update()

    def __plug_in_new_molecule(self, molecule: vtkMolecule) -> None:
        self.__molecule = molecule
        self.__molecule_version = uuid1()
        self.__filter.SetInputData(self.__molecule)
        self.__style.molecule = self.__molecule
        self.__labels.set_molecule(self.__molecule)
        if not self.__disable_modification and \
                self.__molecule.GetNumberOfAtoms() != len(self.__arrow_filter_list):
            self.__create_force_arrows()
        self._atom_selection.set_molecule(self.__molecule)
        self._qm_selection.set_molecule(self.__molecule)
        self.set_selection(self.__selected_atoms)
        if self.__haptic_client is not None:
            self.__haptic_client.update_molecule(self.__molecule)
        if self.__must_show_depth_view():
            self.__depth_molecule_producer.SetOutput(self.__molecule)
        self.__update_atomic_charge_display()
        self.__render_molecule()

    @staticmethod
    def __create_selector(
        renderer: vtkRenderer, tolerance: float = 1e-6
    ) -> vtkSelectVisiblePoints:
        """
        Creates an instance of a class that selects visible points.
        """
        selector = vtkSelectVisiblePoints()
        selector.SetRenderer(renderer)
        selector.SetTolerance(tolerance)

        return selector

    def __add_layout(self) -> None:
        """
        Adds a zero margin layout with the viewer,
        """
        layout = QHBoxLayout(self)
        layout.setMargin(0)
        layout.addWidget(self.__interactor)
        self.setLayout(layout)

    def __load_from_atom_collection(
        self, atom_collection: Any, bond_order_collection: Union[Any, None] = None
    ) -> None:
        """
        Loads a molecule from an atom collection and an optional bond order collection.
        If no bond order collection is given, it is constructed based on distances.
        """
        self.__plug_in_new_molecule(atom_collection_to_molecule(atom_collection))
        if bond_order_collection is None:
            self.__set_bonds_based_on_distances(atom_collection)
        else:
            self.set_bonds(bond_order_collection)
        self.__update_selector()

    def __load_from_file(self, file: Path) -> None:
        """
        Loads a molecule from a file supported by scine_utilities
        """
        traj = None
        ending = file.suffix
        if ending == ".bin":
            traj = utils.io.read_trajectory(utils.io.TrajectoryFormat.Binary, str(file))
        elif ending == ".xyz":
            traj = utils.io.read_trajectory(utils.io.TrajectoryFormat.Xyz, str(file))
        if traj is not None:
            if len(traj) == 1:
                # just single frame, load as simple molecule
                atoms = utils.AtomCollection(traj.elements, traj[0])
                bonds = utils.BondOrderCollection()
            else:
                # display trajectory and then leave (generate no molecule)
                # but second window gives chance to save specific frame
                from scine_heron.molecule.molecule_video import MainVideo
                self.dialog = MainVideo(traj)
                self.dialog.show()
                return
        else:
            atoms, bonds = utils.io.read(str(file))
        if bonds.empty():
            self.__load_from_atom_collection(atoms, None)
        else:
            self.__load_from_atom_collection(atoms, bonds)

    def __set_bonds_based_on_distances(self, atom_collection: Any) -> None:
        bonds = utils.BondDetector.detect_bonds(atom_collection)
        self.set_bonds(bonds)
        self.__render_molecule()

    def set_bonds(self, bond_order_collection: Any) -> None:
        n = bond_order_collection.get_system_size()
        for i in range(n - 1):
            for j in range(i + 1, n):
                bond_id = self.__molecule.GetBondId(i, j)
                v = bond_order_collection.get_order(i, j)
                # vtk only supports integer orders
                self.__molecule.SetBondOrder(bond_id, int(round(v)))

    def __update_selector(self) -> None:
        """
        Updates the world tolerance of the class instance that
        selects the labels to display. Labels are located at the centers of atoms,
        and therefore are always hidden by the atom itself. By providing a
        tolerance which is approximately/at least the radius of the atom,
        we show those labels nonetheless.
        """
        self.__selector.SetToleranceWorld(
            self.__molecule_mapper.GetAtomicRadiusScaleFactor()
            * maximum_vdw_radius(self.__molecule)
        )

    def __update_molecule_style_if_necessary(self) -> None:
        """
        Updates the molecule style using from the status manager if the style
        has not been changed by displaying custom atom radii.
        """
        if (
            self.__molecule_mapper.GetAtomicRadiusType()
            == vtkMoleculeMapper.CustomArrayRadius
        ):
            return
        self.__update_molecule_style()

    def __update_molecule_style(self) -> None:
        """
        Change molecule style.
        """
        if self.__settings_status_manager.molecule_style == MoleculeStyle.BallAndStick:
            self.__molecule_mapper.UseBallAndStickSettings()
        elif self.__settings_status_manager.molecule_style == MoleculeStyle.VDWSpheres:
            self.__molecule_mapper.UseVDWSpheresSettings()
        elif (
            self.__settings_status_manager.molecule_style
            == MoleculeStyle.LiquoriceStick
        ):
            self.__molecule_mapper.UseLiquoriceStickSettings()
        elif self.__settings_status_manager.molecule_style == MoleculeStyle.Fast:
            self.__molecule_mapper.UseFastSettings()
        elif (
            self.__settings_status_manager.molecule_style
            == MoleculeStyle.PartialCharges
        ):
            self.__molecule_mapper.UseBallAndStickSettings()
        elif (
            self.__settings_status_manager.molecule_style
            == MoleculeStyle.ForceArrow
        ):
            self.__molecule_mapper.UseBallAndStickSettings()
        else:
            assert False, "Unknown molecule style type."
        self.__update_selector()
        self.__render_molecule()

    def __update_labels_style(self) -> None:
        """
        Change labels style.
        """
        self.__labels.display_labels_style(self.__settings_status_manager.labels_style)
        self.__render_molecule()

    def provide_data(self) -> vtkMolecule:
        """
        Returns the currently displayed molecule.
        """
        return self.__molecule_mapper.GetInput()

    def set_calc_gradient_in_loop(self, calc_gradient_in_loop: bool) -> None:
        """
        Set calc_gradient_in_loop flag in MoleculeInteractorStyle instance.
        """
        return self.__style.set_calc_gradient_in_loop(calc_gradient_in_loop)

    def display_array_via_colors(self, array: vtkDoubleArray) -> None:
        """
        Display the given array via atom colors. The array needs to contain
        exactly one scalar per atom.
        """
        assert (
            array.GetNumberOfValues() == self.__molecule.GetNumberOfAtoms()
        ), "The array needs to provide one scalar per atom."

        self.__display_colorbar_with_range(array.GetValueRange())
        self.__display_colors(self.__colorbar.GetLookupTable().MapScalars(array, 0, 0))

    def __display_colorbar_with_range(self, value_range: Tuple[float, float]) -> None:
        """
        Adds a colorbar with the given range to the display.
        """
        self.__colorbar.SetLookupTable(
            create_symmetric_color_transfer_function(value_range)
        )
        self.__renderer.AddActor2D(self.__colorbar)

    def __display_colors(self, colors: vtkUnsignedCharArray) -> None:
        """
        Display the given colors per atom. The array needs to contain
        exactly one color per atom.
        """
        assert (
            colors.GetNumberOfTuples() == self.__molecule.GetNumberOfAtoms()
        ), "The array needs to provide one color per atom."

        # we need to control the color mapping, otherwise the mapper uses a
        # color map suitable for atomic numbers
        self.__molecule_mapper.SetMapScalars(False)

        colors.SetName("scine_colors")
        self.__molecule.GetVertexData().AddArray(colors)
        self.__molecule_mapper.SetInputArrayToProcess(
            0, 0, 0, vtkDataObject.FIELD_ASSOCIATION_VERTICES, "scine_colors"
        )

    def display_atomic_number_colors(self) -> None:
        """
        Switches the colors back to the default atomic number based colors.
        """
        self.__renderer.RemoveActor2D(self.__colorbar)
        self.__molecule_mapper.SetMapScalars(True)
        self.__molecule_mapper.SetInputArrayToProcess(
            0,
            0,
            0,
            vtkDataObject.FIELD_ASSOCIATION_VERTICES,
            self.__molecule.GetAtomicNumberArrayName(),
        )
        self.__molecule.GetVertexData().RemoveArray("scine_colors")

    def display_array_via_labels(self, array: vtkDoubleArray) -> None:
        """
        Display the given values as labels per atom.
        """
        assert (
            array.GetNumberOfValues() == self.__molecule.GetNumberOfAtoms()
        ), "The array needs to provide one scalar per atom."

        labels = vtkDoubleArray()
        labels.DeepCopy(array)
        labels.SetName("scine_labels")
        self.__molecule.GetVertexData().AddArray(labels)

    def display_style_labels(self) -> None:
        """
        Display the labels per atom as defined by the labels style.
        """
        self.__molecule.GetVertexData().RemoveArray("scine_labels")

    def display_array_via_radii(self, values: vtkDoubleArray) -> None:
        """
        Display the given radii per atom. The radii are rescaled to [0, m], where
        m is the maximum VDW radius of the molecule multiplied by the default atomic
        radius scale factor of the vtkMoleculeMapper.
        """
        assert (
            values.GetNumberOfValues() == self.__molecule.GetNumberOfAtoms()
        ), "The array needs to provide one radius per atom."

        radii = rescale_to_range(
            values,
            (
                0,
                maximum_vdw_radius(self.__molecule)
                * vtkMoleculeMapper().GetAtomicRadiusScaleFactor(),
            ),
        )
        radii.SetName("radii")
        self.__molecule.GetVertexData().AddArray(radii)
        self.__molecule_mapper.SetAtomicRadiusTypeToCustomArrayRadius()

    def display_atomic_number_radii(self) -> None:
        """
        Switches the radii back to the default atomic number based radii.
        """
        self.__update_molecule_style()

    def __render_molecule(self) -> None:
        """
        Render molecule.
        """
        self.__renderer.GetRenderWindow().Render()

    def add_atom_default_position(self, atomic_number: int) -> vtkMolecule:
        """
        Add an atom in a position given by the structural completion functions.
        """
        if len(self.__selected_atoms) == 0 and self.__molecule.GetNumberOfAtoms() > 0:
            self.__settings_status_manager.error_message = (
                "No selected atom, cannot determine position for new atom."
            )
            return

        atoms_to_consider = list(self.__selected_atoms)
        mol = self.__filter.GetOutput()
        for bond_id in range(mol.GetNumberOfBonds()):
            bond = mol.GetBond(bond_id)
            if bond.GetBeginAtomId() == atoms_to_consider[0]:
                atoms_to_consider.append(bond.GetEndAtomId())
            if bond.GetEndAtomId() == atoms_to_consider[0]:
                atoms_to_consider.append(bond.GetBeginAtomId())

        new_molecule = emf.build_molecule_with_new_atom_structural_completion(
            self.__molecule,
            atomic_number,
            atoms_to_consider,
            self.__settings_status_manager,
        )

        self.__plug_in_new_molecule(new_molecule)

    def has_atoms(self) -> bool:
        return bool(self.__molecule.GetNumberOfAtoms() > 0)

    def remove_selected_atoms(self) -> None:
        new_molecule = emf.build_molecule_without_removed_atoms(
            self.__molecule, self.__selected_atoms
        )
        self.__selected_atoms = []
        if not self.__disable_modification:
            self.set_selection(self.__selected_atoms)
        self.__plug_in_new_molecule(new_molecule)

    def append_selection(self, atoms: Optional[List[int]]) -> None:
        """
        Only works if a selection has already been made because of a round-about way to deselect atoms.
        """
        if len(self.__selected_atoms) == 0:
            return
        if atoms:
            for atom in atoms:
                if atom is None or atom in self.__selected_atoms:
                    return
                self.__selected_atoms.append(atom)
        if not self.__disable_modification:
            self.set_selection(self.__selected_atoms)

    def get_atom_collection(self) -> utils.AtomCollection:
        return molecule_to_atom_collection(self.__molecule)

    def get_energy(self, seconds=0.5) -> Optional[float]:
        if self.__energy_status_manager is None:
            return None
        return self.__energy_status_manager.get_latest_energy(seconds)

    def single_point_calculation(self) -> CustomResult:
        if self.__molecule.GetNumberOfAtoms() == 0:
            self.__settings_status_manager.error_message = "Cannot carry out calculation with no loaded molecule"
            return CustomResult()

        # we simply want to create a separate instance of a Scine Calculator to execute a single point calculation
        # basic logic is contained in this function
        def _impl(calculator_args: Tuple[str, str],
                  calculator_settings: Dict[str, Any],
                  bond_display: str,
                  molecule_version: int,
                  molecule: vtkMolecule,
                  results: ListProxy) -> None:

            method_family, program = calculator_args
            calculation_client = ScineCalculatorWrapper(
                method_family, program, hessian_required=False, settings=calculator_settings,
            )
            atoms = molecule_to_atom_collection(molecule)
            try:
                calculation_client.set_bond_orders_flag(bond_display != "distance")
                calculation_client.update_system(molecule_version, [str(e) for e in atoms.elements], atoms.positions)
                # Carry out calculations
                result = calculation_client.calculate_custom_result()
            except RuntimeError as error:
                result = CustomResult()
                result.error_msg = str(error)
            result.make_pickleable()
            results.append(result)

        # we do this ugly multiprocessing hack, because the following scenario is unsafe:
        # creating the first instance of a calculator here, which may load openmpi libraries if available
        # destroying the calculator
        # forking the process that did this later for continuous interactive updates
        # constructing a new calculator that again utilizes openmpi
        # because some references to the originally loaded libraries may still be in memory,
        # this can lead to undefined behavior
        # we avoid this by forking here and doing the single point calculation in a separate process
        with multiprocessing.Manager() as manager:
            default = CustomResult()
            default.make_pickleable()
            results: ListProxy = manager.list([default])

            p = multiprocessing.Process(target=_impl,
                                        args=(self.__settings_status_manager.get_calculator_args(),
                                              self.__settings_status_manager.get_calculator_settings(),
                                              self.__settings_status_manager.bond_display,
                                              self.__molecule_version,
                                              self.__molecule,
                                              results
                                              )
                                        )
            p.start()
            p.join()
            if len(results) <= 1:
                return default
            result = results[1]
            if result is not None:
                if result.error_msg:
                    self.__settings_status_manager.error_message = result.error_msg
                elif result.info_msg:
                    self.__settings_status_manager.info_message = result.info_msg
            else:
                result = default
            return result

    def get_selection(self) -> List[int]:
        return self._atom_selection.get_selection_data()

    def set_selection(self, selection: List[int], color: Optional[Tuple[float, float, float]] = None):
        if color is None:
            self._atom_selection.back_to_default_color()
        else:
            self._atom_selection.change_color(*color)
        self._atom_selection.set_selection(selection)
        self.__selected_atoms = selection
        self.__render_molecule()

    def make_indices_opaque(self, selected_indices: List[int],
                            _exclude_mode: bool = False,
                            _opacity_percent: float = 0.5):
        self._qm_selection.set_selection(selected_indices)
        """
        # we need to control the color mapping, otherwise the mapper uses a
        # color map suitable for atomic numbers
        self.__molecule_mapper.SetMapScalars(False)
        n = self.__molecule.GetNumberOfAtoms()
        atomic_numbers_data = \
            self.__molecule.GetVertexData().GetArray(self.__molecule.GetAtomicNumberArrayName())
        table = vtkPeriodicTable()
        colors = vtkUnsignedCharArray()
        colors.SetName("opacity")
        colors.SetNumberOfComponents(4)
        colors.SetNumberOfTuples(n)
        for i in range(n):
            rgb_percent = table.GetDefaultRGBTuple(int(atomic_numbers_data.GetTuple(i)[0]))
            in_indices: bool = i in selected_indices
            if in_indices:
                this_opacity = 1 if exclude_mode else opacity_percent
            else:
                this_opacity = opacity_percent if exclude_mode else 1
            colors.SetTuple(i, [p * 255 for p in list(rgb_percent) + [this_opacity]])
        self.__molecule.GetVertexData().AddArray(colors)
        self.__molecule_mapper.SetInputArrayToProcess(
            0, 0, 0, vtkDataObject.FIELD_ASSOCIATION_VERTICES, "opacity"
        )
        self.__render_molecule()
        """

    def remove_opacity(self) -> None:
        self._qm_selection.set_selection([])
        """
        self.__molecule_mapper.SetMapScalars(True)
        self.__molecule_mapper.SetInputArrayToProcess(
            0,
            0,
            0,
            vtkDataObject.FIELD_ASSOCIATION_VERTICES,
            self.__molecule.GetAtomicNumberArrayName(),
        )
        self.__molecule.GetVertexData().RemoveArray("opacity")
        self.__render_molecule()
        """

    def __create_force_arrows(self):
        n_atoms = self.__molecule.GetNumberOfAtoms()
        self.__arrow_filter_list = []
        for _id in range(n_atoms):
            force_vector = vtkArrowSource()
            force_vector_mapper = vtkPolyDataMapper()
            transformFilter = vtkTransformPolyDataFilter()
            transformFilter.SetInputConnection(force_vector.GetOutputPort())
            transform = vtkTransform()
            transformFilter.SetTransform(transform)
            transform.Scale(0.0, 0.0, 0.0)
            force_vector_mapper.SetInputConnection(transformFilter.GetOutputPort())
            force_vector_actor = vtkActor()
            force_vector_actor.SetMapper(force_vector_mapper)
            self.__renderer.AddActor(force_vector_actor)
            self.__arrow_filter_list.append(transformFilter)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
__copyright__ = """ This code is licensed under the 3-clause BSD license.
Copyright ETH Zurich, Department of Chemistry and Applied Biosciences, Reiher Group.
See LICENSE.txt for details.
"""

import numpy as np
from typing import List, TypeVar

Position = TypeVar("Position")


def collision(
    radius1: float, position1: Position, radius2: float, position2: Position
) -> bool:
    """
    Checks whether 2 spheres collide.
    """
    res = bool(
        ((np.array(position1) - np.array(position2)) ** 2).sum()
        < (radius1 + radius2) ** 2
    )

    return res


def collision_multiple(
    radius: float, position: Position, radii: List[float], positions: List[Position]
) -> bool:
    """
    Checks whether a sphere collides with any sphere in a list.
    """
    return any(collision(radius, position, r, p) for r, p in zip(radii, positions))



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
__copyright__ = """ This code is licensed under the 3-clause BSD license.
Copyright ETH Zurich, Department of Chemistry and Applied Biosciences, Reiher Group.
See LICENSE.txt for details.
"""
"""
Provides the GroupedComboBox class.
"""

from PySide2.QtCore import QObject
from PySide2.QtGui import QStandardItem
from scine_heron.containers.combo_box import BaseBox
from scine_heron.chemoton.class_searcher import ChemotonClassSearcher


class GroupedComboBox(BaseBox):
    """
    A combo box that can be constructed with a ChemotonClassSearcher to include all its
    held classes as selectable items and split them based on its module information.
    """

    def __init__(self, parent: QObject, class_searcher: ChemotonClassSearcher, add_none: bool = False) -> None:
        """
        Construct the widget based on the class searcher

        Parameters
        ----------
        parent : QObject
            The parent widget
        class_searcher : ChemotonClassSearcher
            The class searcher holding the classes that are selectable
        add_none : bool, optional
            If 'None' should be added as an option to the combo box, by default False
        """
        BaseBox.__init__(self, parent=parent)

        self.gear_searcher = class_searcher
        child_buffer = ' ' * 3
        child_name = str()  # use for gear name selection.

        if add_none:
            parent_item = QStandardItem("none")
            parent_item.setEnabled(False)
            self.model().appendRow(parent_item)
            child_name = child_buffer + "none"
            child_item = QStandardItem(child_name)
            self.model().appendRow(child_item)

        for module in sorted(self.gear_searcher.module_to_class.keys()):
            parent_item = QStandardItem(module)
            parent_item.setEnabled(False)
            self.model().appendRow(parent_item)

            for item_name in self.gear_searcher.module_to_class[module]:
                child_name = child_buffer + item_name
                child_item = QStandardItem(child_name)
                self.model().appendRow(child_item)

        self.setCurrentText(child_name)



__copyright__ = """ This code is licensed under the 3-clause BSD license.
Copyright ETH Zurich, Department of Chemistry and Applied Biosciences, Reiher Group.
See LICENSE.txt for details.
"""

from typing import TYPE_CHECKING, Any

if TYPE_CHECKING:
    Signal = Any
else:
    pass
import numpy as np
from PySide2.QtCore import QObject, Signal


class SignalHandler(QObject):
    open_entanglement_widget_signal = Signal()
    update_entanglement_plot_signal = Signal(np.ndarray, np.ndarray, list)
    update_mo_diagram = Signal()
    change_iso_value = Signal(float)

    open_molecule_widget_signal = Signal()
    load_xyz_file_signal = Signal(str)
    load_molden_file_signal = Signal(str)
    load_orbital_groups_file_signal = Signal(str)
    view_orbital = Signal(int)

    start_initial_orbital_calculation_signal = Signal()
    start_initial_dmrg_calculation_signal = Signal()
    start_final_calculation_signal = Signal()

    toggle_file_tree = Signal()

    def __init__(self, parent=None):
        if parent is not None:
            super().__init__(parent)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
__copyright__ = """ This code is licensed under the 3-clause BSD license.
Copyright ETH Zurich, Department of Chemistry and Applied Biosciences, Reiher Group.
See LICENSE.txt for details.
"""

from multiprocessing import Process, Manager
from multiprocessing.managers import SyncManager
from time import sleep
import typing
import socket
import numpy as np

import scine_utilities as su
from scine_heron.calculators.calculator import ScineCalculatorWrapper
from .mediator_potential import MediatorPotential
from .system import System
from .clientserver import recv_data, send_data


def convert_positions_to_bohr(
        molecule: typing.List[typing.List[typing.Union[float, str]]]
) -> np.ndarray:
    """
    Conversion to bohr (a.u.)
    """
    return np.array([xyz for _, xyz in molecule], dtype=np.float64) * su.BOHR_PER_ANGSTROM


def get_results_to_send(
    mediator_potential: MediatorPotential,
    positions: np.ndarray,
    molecule_version: int,
    natoms: int,
) -> typing.Tuple[
    float,
    np.ndarray,
    typing.Optional[typing.List[float]],
    typing.Optional[np.ndarray],
    typing.Optional[str],
    typing.Dict[str, typing.Any],
]:
    """
    Computes the energy and the gradient
    given the mediator potential and the positions.
    In case something goes wrong, returns sensible defaults.
    """

    if (
            mediator_potential is None
            or mediator_potential.molecule_version != molecule_version
    ):
        energy, gradient, charges, bond_orders, molden_input, settings = (
            0.0,
            np.zeros((natoms, 3)),
            None,
            None,
            None,
            {},
        )
    else:
        energy = mediator_potential.get_energy(positions)
        gradient = mediator_potential.get_gradients(positions)
        charges = mediator_potential.get_atomic_charges()
        bond_orders = mediator_potential.bond_orders
        if bond_orders is not None:
            # nd.array is not JSON serializable, so we convert it to a list
            bond_orders = bond_orders.tolist()
        molden_input = mediator_potential.molden_input
        settings = mediator_potential.settings
    return energy, gradient, charges, bond_orders, molden_input, settings


def server(
        shared_data: typing.MutableMapping[str, typing.Any],
        lock: typing.ContextManager[None],
) -> None:
    """
    Connects to the client (GUI)
    Receives molecule, settings i.e. updates system
    Sends gradients and energy derived from mediator potential
    """
    s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    s.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
    s.bind(("127.0.0.1", 55145))
    s.listen()
    active_mediator_potential = True
    saved_mediator_potential: typing.Optional[MediatorPotential] = None

    last_error = ""
    while True:
        conn, _ = s.accept()
        with conn:
            stop_signal, molecule_version, molecule, calculator_args, settings, mediator_potential_signal, \
                bond_orders_type = recv_data(conn)
            shared_data["stop_signal"] = stop_signal

            if stop_signal:
                break

            if mediator_potential_signal is not None:
                # we received signal to change set switch
                active_mediator_potential = mediator_potential_signal
                shared_data["mediator_potential_signal"] = mediator_potential_signal

            positions = convert_positions_to_bohr(molecule)
            atom_symbols = [atom[0] for atom in molecule]
            with lock:
                shared_data["system"] = System(
                    molecule_version=molecule_version,
                    positions=positions,
                    atom_symbols=atom_symbols,
                    calculator_args=calculator_args,
                    settings=settings,
                )
                shared_data["bond_orders_type"] = bond_orders_type
                mediator_potential = shared_data.get("mediator_potential", None)
                error_msg = shared_data.get("error_msg", last_error)
                info_msg = shared_data.get("info_msg", "")

                if mediator_potential is not None and not active_mediator_potential:
                    # result evaluation relies on the fact that it always happens over the mediator potential class
                    # independent of whether we have a real finished calculation or the mediator
                    # to disable it, we rely on the fact that it was not None, so we once got a result
                    # we evaluate energy, gradients and now remove the class from the shared data
                    # only once a new real calculation was finished, we have a mediator potential again, until then
                    # we just send zeros because we get a None from the shared data again
                    saved_mediator_potential = mediator_potential  # back up in case it is activated again
                    del shared_data["mediator_potential"]
                if mediator_potential is None and active_mediator_potential and saved_mediator_potential is not None:
                    mediator_potential = saved_mediator_potential

            (
                energy,
                gradient,
                charges,
                bond_orders,
                molden_input,
                settings,
            ) = get_results_to_send(
                mediator_potential, positions, molecule_version, len(molecule)
            )

            send_data([energy, gradient.tolist(), charges, bond_orders, molden_input,
                       error_msg, info_msg, settings, mediator_potential_signal],
                      conn,
                      )


def electronic_structure_calculation(
        shared_data: typing.MutableMapping[str, typing.Any],
        lock: typing.ContextManager[None],
) -> None:
    """
    Prepares calculator for Sparrow calculations
    Updates mediator potential whenever new results are available
    """
    calculation_client = None
    prev_mediator_signal = True
    while True:
        try:
            with lock:
                system = shared_data.get("system", None)
                stop_signal = shared_data.get("stop_signal", False)
                mediator_signal = shared_data.get("mediator_potential_signal", True)
                bond_orders_type = shared_data.get("bond_orders_type", "distance")
            if stop_signal:
                break

            if system is None:
                continue

            if calculation_client is None:
                calculation_client = ScineCalculatorWrapper(  # type: ignore
                    *system.calculator_args,
                    hessian_required=mediator_signal,
                    settings=system.settings,
                )

            if prev_mediator_signal != mediator_signal:
                calculation_client.set_hessian_flag(mediator_signal)
                prev_mediator_signal = mediator_signal

            error_msg = ""
            mediator_potential = None
            try:
                calculation_client.set_bond_orders_flag(bond_orders_type != "distance")
                calculation_client.update_system(
                    system.molecule_version, system.atom_symbols, system.positions, system.settings
                )
                # Perform calculations
                result = calculation_client.calculate_custom_result()
            except RuntimeError as error:
                error_msg = str(error)
            else:
                if result:
                    mediator_potential = MediatorPotential(system.molecule_version, result)
                    with lock:
                        shared_data["info_msg"] = result.info_msg
                elif result is not None:
                    error_msg = result.error_msg
            with lock:
                if mediator_potential:
                    shared_data["mediator_potential"] = mediator_potential
                shared_data["error_msg"] = error_msg
        except BrokenPipeError:
            break


def run_server(stop_signal: typing.Callable) -> None:
    with Manager() as manager:
        assert isinstance(manager, SyncManager)

        shared_data: typing.MutableMapping[str, typing.Any] = manager.dict()
        lock = manager.Lock()

        server_process = Process(target=server, args=(shared_data, lock))
        server_process.start()

        calculation_process = Process(
            target=electronic_structure_calculation, args=(shared_data, lock)
        )
        calculation_process.start()

        while all(p.is_alive() for p in [server_process, calculation_process]):
            sleep(0.1)

        if any(p.is_alive() for p in [server_process, calculation_process]):
            sleep(0.5)
            try:
                stop_signal()
            except ConnectionRefusedError:
                pass

        if calculation_process.is_alive():
            server_process.join()
            calculation_process.join()
        else:
            calculation_process.join()
            server_process.join()



__copyright__ = """ This code is licensed under the 3-clause BSD license.
Copyright ETH Zurich, Department of Chemistry and Applied Biosciences, Reiher Group.
See LICENSE.txt for details.
"""

from typing import List

from PySide2.QtWidgets import QGridLayout, QLabel, QPushButton, QWidget


class OrbitalOptionsWidget(QWidget):
    def __init__(self, parent: QWidget) -> None:
        QWidget.__init__(self, parent)
        self.__layout = QGridLayout()
        self.buttons: List[QPushButton] = []
        self.setLayout(self.__layout)

    def update(self):
        self.buttons = []
        button_grid = [[]]
        last = None
        diff = 10.0
        row = 0
        energies = self.parent().options.orbital_energies
        occupations = self.parent().options.occupations
        for i, e in enumerate(reversed(energies)):
            # print(i, e)
            x = len(energies) - i - 1
            string = "{:n}".format(x - 1)
            # print(string)
            button = QPushButton(string)
            button.setCheckable(True)
            button.setMinimumWidth(60)
            button.setMaximumWidth(60)
            button.setMinimumHeight(20)
            button.setMaximumHeight(20)
            self.buttons.append(button)
            if last is not None:
                diff = abs(e - last)
            if diff > 1e-4:
                row += 1
                button_grid.append([button])
            else:
                button_grid[row].append(button)
            last = e
        max_cols = 4
        r = 0
        counter = 0
        printed_gap_info = False
        for row in button_grid:
            modulo_modifier = 0
            for c, col in enumerate(row):
                if occupations[len(energies) - counter - 1] and not printed_gap_info:
                    printed_gap_info = True
                    r += 1
                    # self.layout.addWidget(QHLine(), r, 0, 1, max_cols)
                    r += 1
                    label = QLabel("HOMO-LUMO Gap")
                    label.setMinimumHeight(20)
                    label.setMaximumHeight(20)
                    self.__layout.addWidget(label, r, 0, 1, max_cols)
                    r += 1
                    # self.layout.addWidget(QHLine(), r, 0, 1, max_cols)
                    r += 1
                    modulo_modifier = c
                mod = (c - modulo_modifier) % max_cols
                if mod == 0:
                    r += 1
                self.__layout.addWidget(col, r, mod, 1, 1)
                counter += 1
            r += 1
            # self.layout.addWidget(QHLine(), r, 0, 1, max_cols)

    def get_checked(self):
        cas = []
        cas_occupation = []
        occupation = self.parent().parent().parent().settings.occupations
        for i, b in enumerate(self.buttons[::-1]):
            if b.isChecked():
                # print(i, occupation[i])
                cas.append(i)
                cas_occupation.append(occupation[i])
        self.parent().parent().parent().settings.cas = cas
        self.parent().parent().parent().settings.cas_occupations = cas_occupation
